from AppKit import NSLevelIndicator, NSLevelIndicatorCell, NSTickMarkAbove, NSTickMarkBelow, \
        NSDiscreteCapacityLevelIndicatorStyle, NSContinuousCapacityLevelIndicatorStyle, \
        NSRatingLevelIndicatorStyle, NSRelevancyLevelIndicatorStyle, NSImage
from vanilla.vanillaBase import VanillaBaseControl

# This control is available in OS 10.4+.
# Cause a NameError if in an earlier OS.
NSLevelIndicator


_tickPositionMap = {
    "above": NSTickMarkAbove,
    "below": NSTickMarkBelow,
}

_levelIndicatorStyleMap = {
    "discrete":   NSDiscreteCapacityLevelIndicatorStyle,
    "continuous": NSContinuousCapacityLevelIndicatorStyle,
    "rating":     NSRatingLevelIndicatorStyle,
    "relevancy":  NSRelevancyLevelIndicatorStyle,
}


class LevelIndicator(VanillaBaseControl):

    """
    A control which shows a value on a linear scale.

    .. image:: /_images/LevelIndicator.png

    ::

        from vanilla import Window, LevelIndicator

        class LevelIndicatorDemo:

             def __init__(self):
                 self.w = Window((200, 68))
                 self.w.discreteIndicator = LevelIndicator(
                            (10, 10, -10, 18), callback=self.levelIndicatorCallback)
                 self.w.continuousIndicator = LevelIndicator(
                            (10, 40, -10, 18), style="continuous",
                            callback=self.levelIndicatorCallback)
                 self.w.open()

             def levelIndicatorCallback(self, sender):
                 print("level indicator edit!", sender.get())

        LevelIndicatorDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing the position
    and size of the level indicator.

    +----------------------------------------+
    | **Standard Dimensions**                |
    +===============================+===+====+
    | *discrete without ticks*      | H | 18 |
    +-------------------------------+---+----+
    | *discrete with minor ticks*   | H | 22 |
    +-------------------------------+---+----+
    | *discrete with major ticks*   | H | 25 |
    +-------------------------------+---+----+
    | *continuous without ticks*    | H | 16 |
    +-------------------------------+---+----+
    | *continuous with minor ticks* | H | 20 |
    +-------------------------------+---+----+
    | *continuous with major ticks* | H | 23 |
    +-------------------------------+---+----+

    **style** The style of the level indicator. The options are:

    +--------------+-------------------+
    | "continuous" | A continuous bar. |
    +--------------+-------------------+
    | "discrete"   | A segmented bar.  |
    +--------------+-------------------+

    **value** The initial value of the level indicator.

    **minValue** The minimum value allowed by the level indicator.

    **maxValue** The maximum value allowed by the level indicator.

    **warningValue** The value at which the filled portions of the
    level indicator should display the warning color.

    **criticalValue** The value at which the filled portions of the
    level indicator should display the critical color.

    **tickMarkPosition** The position of the tick marks in relation
    to the level indicator. The options are:

    +---------+
    | "above" |
    +---------+
    | "below" |
    +---------+

    **minorTickMarkCount** The number of minor tick marcks to be displayed
    on the level indicator. If *None* is given, no minor tick marks will be displayed.

    **majorTickMarkCount** The number of major tick marcks to be displayed on the level
    indicator. If *None* is given, no major tick marks will be displayed.

    **callback** The method to be called when the level indicator has been edited.
    If no callback is given, the level indicator will not be editable.
    """

    nsLevelIndicatorClass = NSLevelIndicator

    def __init__(self, posSize, style="discrete",
                    value=5, minValue=0, maxValue=10, warningValue=None, criticalValue=None,
                    tickMarkPosition=None, minorTickMarkCount=None, majorTickMarkCount=None,
                    callback=None):
        self._setupView(self.nsLevelIndicatorClass, posSize, callback=callback)
        self._nsObject.cell().setLevelIndicatorStyle_(_levelIndicatorStyleMap[style])
        self._nsObject.setMinValue_(minValue)
        self._nsObject.setMaxValue_(maxValue)
        self._nsObject.setFloatValue_(value)
        if warningValue is not None:
            self._nsObject.setWarningValue_(warningValue)
        if criticalValue is not None:
            self._nsObject.setCriticalValue_(criticalValue)
        if tickMarkPosition is not None:
            self._nsObject.setTickMarkPosition_(_tickPositionMap[tickMarkPosition])
        if minorTickMarkCount is not None:
            self._nsObject.setNumberOfTickMarks_(minorTickMarkCount)
        if majorTickMarkCount is not None:
            self._nsObject.setNumberOfMajorTickMarks_(majorTickMarkCount)
        if callback is None:
            self._nsObject.setEnabled_(False)

    def getNSLevelIndicator(self):
        """
        Return the `NSLevelIndicator`_ that this object wraps.

        .. _NSLevelIndicator: https://developer.apple.com/documentation/appkit/nslevelindicator?language=objc
        """
        return self._nsObject

    def set(self, value):
        """
        Set the value of the level indicator.
        """
        self._nsObject.setFloatValue_(value)

    def get(self):
        """
        Get the value of the level indicator.
        """
        return self._nsObject.floatValue()

    def setMinValue(self, value):
        """
        Set the minimum value of the level indicator.
        """
        self._nsObject.setMinValue_(value)

    def getMinValue(self):
        """
        Get the minimum value of the level indicator.
        """
        return self._nsObject.minValue()

    def setMaxValue(self, value):
        """
        Set the maximum of the level indicator.
        """
        self._nsObject.setMaxValue_(value)

    def getMaxValue(self):
        """
        Get the maximum of the level indicator.
        """
        return self._nsObject.maxValue()

    def setWarningValue(self, value):
        """
        Set the warning value of the level indicator.
        """
        self._nsObject.setWarningValue_(value)

    def getWarningValue(self, value):
        """
        Get the warning value of the level indicator.
        """
        return self._nsObject.warningValue()

    def setCriticalValue(self, value):
        """
        Set the critical value of the level indicator.
        """
        self._nsObject.setCriticalValue_(value)

    def getCriticalValue(self, value):
        """
        Get the critical value of the level indicator.
        """
        return self._nsObject.criticalValue()


def LevelIndicatorListCell(style="discrete",
        minValue=0, maxValue=10, warningValue=None, criticalValue=None,
        imagePath=None, imageNamed=None, imageObject=None):

    """
    An object that displays a level indicator in a List column.

    .. image:: /_images/LevelIndicatorListCell.png

    .. note::
       This object should only be used in the *columnDescriptions* argument
       during the construction of a List.

    ::

        from vanilla import Window, List, LevelIndicatorListCell

        class LevelIndicatorListCellDemo:

             def __init__(self):
                 self.w = Window((340, 140))
                 items = [
                     {"discrete": 3, "continuous": 4, "rating": 1, "relevancy": 9},
                     {"discrete": 8, "continuous": 3, "rating": 5, "relevancy": 5},
                     {"discrete": 3, "continuous": 7, "rating": 3, "relevancy": 4},
                     {"discrete": 2, "continuous": 5, "rating": 4, "relevancy": 7},
                     {"discrete": 6, "continuous": 9, "rating": 3, "relevancy": 2},
                     {"discrete": 4, "continuous": 0, "rating": 6, "relevancy": 8},
                 ]
                 columnDescriptions = [
                     {"title": "discrete",
                     "cell": LevelIndicatorListCell(style="discrete", warningValue=7, criticalValue=9)},
                     {"title": "continuous",
                     "cell": LevelIndicatorListCell(style="continuous", warningValue=7, criticalValue=9)},
                     {"title": "rating",
                     "cell": LevelIndicatorListCell(style="rating", maxValue=6)},
                     {"title": "relevancy",
                     "cell": LevelIndicatorListCell(style="relevancy")},
                 ]
                 self.w.list = List((0, 0, -0, -0), items=items,
                                columnDescriptions=columnDescriptions)
                 self.w.open()

        LevelIndicatorListCellDemo()

    **style** The style of the level indicator. The options are:

    +--------------+-----------------------------------------+
    | "continuous" | A continuous bar.                       |
    +--------------+-----------------------------------------+
    | "discrete"   | A segmented bar.                        |
    +--------------+-----------------------------------------+
    | "rating"     | A row of stars. Similar to the rating   |
    |              | indicator in iTunes.                    |
    +--------------+-----------------------------------------+
    | "relevancy"  | A row of lines. Similar to the search   |
    |              | result relevancy indicator in Mail.     |
    +--------------+-----------------------------------------+

    **minValue** The minimum value allowed by the level indicator.

    **maxValue** The maximum value allowed by the level indicator.

    **warningValue** The value at which the filled portions of the
    level indicator should display the warning color. Applies only to
    discrete and continuous level indicators.

    **criticalValue** The value at which the filled portions of the
    level indicator should display the critical color. Applies only to
    discrete and continuous level indicators.
    """
    cell = NSLevelIndicatorCell.alloc().init()
    cell.setLevelIndicatorStyle_(_levelIndicatorStyleMap[style])
    cell.setMinValue_(minValue)
    cell.setMaxValue_(maxValue)
    if warningValue is not None:
        cell.setWarningValue_(warningValue)
    if criticalValue is not None:
        cell.setCriticalValue_(criticalValue)
    if imagePath is not None:
        image = NSImage.alloc().initWithContentsOfFile_(imagePath)
    elif imageNamed is not None:
        image = NSImage.imageNamed_(imageNamed)
    elif imageObject is not None:
        image = imageObject
    if imageObject is not None:
        cell.setImage_(image)
    return cell



from AppKit import NSDocument
from PyObjCTools import AppHelper
from tinyTextEditorDocumentWindow import TinyTextEditorDocumentWindow
from io import open


class TinyTextEditorDocument(NSDocument):
    
    def init(self):
        self = super(TinyTextEditorDocument, self).init()
        self.vanillaWindowController = TinyTextEditorDocumentWindow()
        self.vanillaWindowController.assignToDocument(self)
        return self        
    
    def readFromFile_ofType_(self, path, tp):
        # refer to the NSDocument reference for information about this method
        f = open(path, 'r', encoding='utf-8')
        text = f.read()
        f.close()
        self.vanillaWindowController.setText(text)
        return True
    
    def writeWithBackupToFile_ofType_saveOperation_(self, fileName, fileType, operation):
        # refer to the NSDocument reference for information about this method
        text = self.vanillaWindowController.getText()
        f = open(fileName, 'w', encoding='utf-8')
        f.write(text)
        f.close()
        return True


if __name__ == "__main__":
    AppHelper.runEventLoop()


import os
import objc
import AppKit
import vanilla
from vanilla.dragAndDrop import dropOperationMap



# ----
# Test
# ----

objc.options.verbose = True

draggingImages = []
colors = [
    (0, 0, 0),
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1),
    (1, 1, 0),
    (0, 1, 1),
    (1, 0, 1),
    (0.5, 1, 0),
    (0, 0.5, 1),
    (1, 1, 1)
]
for color in colors:
    image = AppKit.NSImage.alloc().initWithSize_((50, 50))
    image.lockFocus()
    r, g, b = color
    AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(r, g, b, 0.8).set()
    path = AppKit.NSBezierPath.bezierPathWithRoundedRect_xRadius_yRadius_(
        ((0, 0), image.size()),
        10,
        10
    )
    path.fill()
    shadow = AppKit.NSShadow.alloc().init()
    shadow.setShadowOffset_((0, -5))
    shadow.setShadowBlurRadius_(5)
    shadow.setShadowColor_(AppKit.NSColor.blackColor())
    shadow.set()
    text = AppKit.NSAttributedString.alloc().initWithString_attributes_(
        "🐲",
        {
            AppKit.NSFontAttributeName : AppKit.NSFont.systemFontOfSize_(40),
            AppKit.NSShadowAttributeName : shadow
        }
    )
    text.drawAtPoint_((3, 3))
    image.unlockFocus()
    draggingImages.append(image)

draggingValues = dict(
    string=[],
    plist=[]
)
for i in range(10):
    letters = "ABCDEFGIJKLMNOPQRSTUVWXYZ"
    draggingValues["string"].append(letters[i])
    draggingValues["plist"].append(
        dict(
            letter=letters[i],
            number=i
        )
    )

class DraggingSourceNSView(AppKit.NSView):

    draggingItemCount = 1
    draggingFormation = "default"
    draggingType = "string"

    def drawRect_(self, rect):
        rect = self.bounds()
        AppKit.NSColor.whiteColor().set()
        AppKit.NSRectFill(rect)
        rect = AppKit.NSInsetRect(rect, 20, 20)
        AppKit.NSColor.greenColor().set()
        AppKit.NSRectFill(rect)

    def mouseDown_(self, event):
        width, height = self.frame().size
        x = (width * 0.5)
        items = []
        for i in range(self.draggingItemCount):
            image = draggingImages[i]
            value = draggingValues[self.draggingType][i]
            y = height * i * 0.1
            location = (x, y)
            item = dict(
                typesAndValues={
                    self.draggingType : value
                },
                image=image,
                location=location
            )
            items.append(item)
        vanilla.startDraggingSession(
            view=self,
            event=event,
            items=items,
            formation=self.draggingFormation
        )


class DraggingSourceView(vanilla.Group):

    nsViewClass = DraggingSourceNSView


titleSpacing = 5
controlSpacing = 10

class Test:

    def __init__(self):
        self.w = vanilla.Window(
            (600, 1),
            "Drag and Drop Test"
        )

        self.makeMainSource()
        self.makeDestViews()
        self.makeDestLists()

        self.w.stack = vanilla.VerticalStackView(
            "auto",
            views=[
                vanilla.TextBox(
                    "auto",
                    "Drag from the green block to one of "
                    "the yellow blocks or lists."
                    "\n"
                    "Drag files from Finder to the file "
                    "yellow block or the list."
                    "\n"
                    "The string list is reorderable."
                ),
                dict(
                    view=self.mainSourceStack,
                    width="fill"
                ),
                dict(
                    view=self.destViewViewStack,
                    height=150,
                    width="fill"
                ),
                dict(
                    view=self.destViewListStack,
                    height=150,
                    width="fill"
                ),
                # vanilla.HorizontalLine("auto"),
                # vanilla.TextBox("auto", "Drag within the list to reorder items."),
            ],
            spacing=20,
            alignment="leading"
        )

        rules = [
            "H:|-[stack]-|",
            "V:|-[stack]-|"
        ]
        self.w.addAutoPosSizeRules(rules)

        self.w.open()

    def makeMainSource(self):
        self.dragSourceTitle = vanilla.TextBox(
            "auto",
            "Drag Details:"
        )
        self.dragSourceCountSlider = vanilla.Slider(
            "auto",
            minValue=1,
            maxValue=10,
            value=1,
            tickMarkCount=10,
            stopOnTickMarks=True,
            callback=self.dragSourceSettingsCallback
        )
        self.dragSourceTypePopUp = vanilla.PopUpButton(
            "auto",
            [
                "string",
                "plist"
            ],
            callback=self.dragSourceSettingsCallback
        )
        self.dragSourceFormationPopUp = vanilla.PopUpButton(
            "auto",
            [
                "default",
                "none",
                "pile",
                "stack",
                "list"
            ],
            callback=self.dragSourceSettingsCallback
        )
        self.dragSourceSettingsStack = vanilla.VerticalStackView(
            "auto",
            views=[
                dict(
                    view=vanilla.TextBox("auto", "Drag Count:"),
                    width="fill",
                    spacing=titleSpacing
                ),
                dict(
                    view=self.dragSourceCountSlider,
                    width="fill"
                ),
                dict(
                    view=vanilla.TextBox("auto", "Drag Type:"),
                    width="fill",
                    spacing=titleSpacing
                ),
                dict(
                    view=self.dragSourceTypePopUp,
                    width="fill"
                ),
                dict(
                    view=vanilla.TextBox("auto", "Drag Formation:"),
                    width="fill",
                    spacing=titleSpacing
                ),
                dict(
                    view=self.dragSourceFormationPopUp,
                    width="fill"
                )
            ],
            spacing=controlSpacing,
            alignment="leading"
        )

        # source

        self.dragSourceView = DraggingSourceView("auto")

        # drop settings

        self.dropOperation = "copy"

        self.dropDestOperationPopUp = vanilla.PopUpButton(
            "auto",
            dropOperationMap.keys(),
            callback=self.dropDestSettingsCallback
        )
        index = list(dropOperationMap).index(self.dropOperation)
        self.dropDestOperationPopUp.set(index)

        self.dropDestOnRowCheckBox = vanilla.CheckBox(
            "auto",
            title="Drop On Row",
            value=False,
            callback=self.dropDestSettingsCallback
        )
        self.dropDestBetweenRowsCheckBox = vanilla.CheckBox(
            "auto",
            title="Drop Between Rows",
            value=True,
            callback=self.dropDestSettingsCallback
        )

        self.dropDestSettingsStack = vanilla.VerticalStackView(
            "auto",
            views=[
                dict(
                    view=vanilla.TextBox("auto", "Drop Mode:"),
                    width="fill",
                    spacing=titleSpacing
                ),
                dict(
                    view=self.dropDestOperationPopUp,
                    width="fill"
                ),
                dict(
                    view=self.dropDestOnRowCheckBox,
                    width="fill",
                    spacing=titleSpacing
                ),
                dict(
                    view=self.dropDestBetweenRowsCheckBox,
                    width="fill",
                    spacing=titleSpacing
                )
            ],
            spacing=controlSpacing,
            alignment="leading"
        )
        self.mainSourceStack = vanilla.HorizontalStackView(
            "auto",
            views=[
                dict(
                    view=self.dragSourceView,
                    width=100
                ),
                dict(
                    view=self.dragSourceSettingsStack,

                ),
                dict(
                    view=self.dropDestSettingsStack,

                )
            ],
            spacing=20,
            alignment="leading"
        )

    def dragSourceSettingsCallback(self, sender):
        count = self.dragSourceCountSlider.get()
        typ = self.dragSourceTypePopUp.get()
        typ = self.dragSourceTypePopUp.getItems()[typ]
        formation = self.dragSourceFormationPopUp.get()
        formation = self.dragSourceFormationPopUp.getItems()[formation]
        view = self.dragSourceView.getNSView()
        view.draggingItemCount = int(count)
        view.draggingFormation = formation
        view.draggingType = typ

    def dropDestSettingsCallback(self, sender):
        operations = self.dropDestOperationPopUp.getItems()
        index = self.dropDestOperationPopUp.get()
        self.dropOperation = operations[index]
        self.stringDestList._allowDropOnRow = self.dropDestOnRowCheckBox.get()
        self.stringDestList._allowDropBetweenRows = self.dropDestBetweenRowsCheckBox.get()
        self.plistDestList._allowDropOnRow = self.dropDestOnRowCheckBox.get()
        self.plistDestList._allowDropBetweenRows = self.dropDestBetweenRowsCheckBox.get()
        self.fileURLDestList._allowDropOnRow = self.dropDestOnRowCheckBox.get()
        self.fileURLDestList._allowDropBetweenRows = self.dropDestBetweenRowsCheckBox.get()

    # Destination Views
    # -----------------

    def makeDestViews(self):
        dropSettings = dict(
            pasteboardTypes=["string"],
            dropCandidateEnteredCallback=self.stringDestViewDropCandidateEnteredCallback,
            dropCandidateCallback=self.stringDestViewDropCandidateCallback,
            dropCandidateEndedCallback=self.stringDestViewDropCandidateEndedCallback,
            dropCandidateExitedCallback=self.stringDestViewDropCandidateExitedCallback,
            performDropCallback=self.stringDestViewPerformDropCallback
        )
        self.stringDestView = vanilla.Group(
            "auto",
            dropSettings=dropSettings
        )
        self.stringDestView.box = vanilla.Box(
            (0, 0, 0, 0),
            fillColor=AppKit.NSColor.yellowColor(),
            borderColor=AppKit.NSColor.greenColor(),
            borderWidth=0,
            margins=0,
            cornerRadius=0
        )
        self.stringDestView.textBox = vanilla.TextBox(
            (20, 20, 0, 0),
            "drop: string"
        )
        self.stringDestView.locationBox = vanilla.Box(
            (0, 0, 2, 2),
            fillColor=AppKit.NSColor.clearColor(),
            borderColor=AppKit.NSColor.redColor(),
            borderWidth=1,
            cornerRadius=10,
            margins=0
        )
        self.stringDestView.locationBox.show(False)

        dropSettings = dict(
            pasteboardTypes=["plist"],
            dropCandidateEnteredCallback=self.plistDestViewDropCandidateEnteredCallback,
            dropCandidateCallback=self.plistDestViewDropCandidateCallback,
            dropCandidateEndedCallback=self.plistDestViewDropCandidateEndedCallback,
            dropCandidateExitedCallback=self.plistDestViewDropCandidateExitedCallback,
            performDropCallback=self.plistDestViewPerformDropCallback
        )
        self.plistDestView = vanilla.Group(
            "auto",
            dropSettings=dropSettings
        )
        self.plistDestView.box = vanilla.Box(
            (0, 0, 0, 0),
            fillColor=AppKit.NSColor.yellowColor(),
            borderColor=AppKit.NSColor.blackColor(),
            margins=0,
            cornerRadius=0
        )
        self.plistDestView.textBox = vanilla.TextBox(
            (20, 20, 0, 0),
            "drop: plist"
        )
        self.plistDestView.locationBox = vanilla.Box(
            (0, 0, 2, 2),
            fillColor=AppKit.NSColor.clearColor(),
            borderColor=AppKit.NSColor.redColor(),
            borderWidth=1,
            cornerRadius=10,
            margins=0
        )
        self.plistDestView.locationBox.show(False)

        dropSettings = dict(
            pasteboardTypes=["fileURL"],
            dropCandidateEnteredCallback=self.fileURLDestViewDropCandidateEnteredCallback,
            dropCandidateCallback=self.fileURLDestViewDropCandidateCallback,
            dropCandidateEndedCallback=self.fileURLDestViewDropCandidateEndedCallback,
            dropCandidateExitedCallback=self.fileURLDestViewDropCandidateExitedCallback,
            performDropCallback=self.fileURLDestViewPerformDropCallback
        )
        self.fileURLDestView = vanilla.Group(
            "auto",
            dropSettings=dropSettings
        )
        self.fileURLDestView.box = vanilla.Box(
            (0, 0, 0, 0),
            fillColor=AppKit.NSColor.yellowColor(),
            borderColor=AppKit.NSColor.blackColor(),
            margins=0,
            cornerRadius=0
        )
        self.fileURLDestView.textBox = vanilla.TextBox(
            (20, 20, 0, 0),
            "drop: files"
        )
        self.fileURLDestView.locationBox = vanilla.Box(
            (0, 0, 2, 2),
            fillColor=AppKit.NSColor.clearColor(),
            borderColor=AppKit.NSColor.redColor(),
            borderWidth=1,
            cornerRadius=10,
            margins=0
        )
        self.fileURLDestView.locationBox.show(False)

        self.destViewViewStack = vanilla.HorizontalStackView(
            "auto",
            views=[
                self.stringDestView,
                self.plistDestView,
                dict(
                    view=vanilla.VerticalLine("auto"),
                    height="fill",
                    width=1
                ),
                self.fileURLDestView
            ],
            spacing=20
        )

    # string

    def stringDestViewDropCandidateEnteredCallback(self, info):
        self.stringDestView.box.setBorderColor(AppKit.NSColor.greenColor())
        self.stringDestView.box.setBorderWidth(20)
        self.stringDestView.locationBox.show(True)
        return self.dropOperation

    def stringDestViewDropCandidateCallback(self, info):
        sender = info["sender"]
        source = info["source"]
        items = info["items"]
        items = sender.getDropItemValues(
            items,
            "string"
        )
        location = info["location"]
        x, y = location
        self.stringDestView.locationBox.setPosSize((x-10, y-10, 20, 20))
        text = (
            f"count={len(items)}",
            f"items={repr(items)}",
            f"source={source.__class__.__name__}"
        )
        text = "\n".join(text)
        self.stringDestView.textBox.set(text)
        return self.dropOperation

    def stringDestViewDropCandidateExitedCallback(self, info):
        self.stringDestView.box.setBorderColor(AppKit.NSColor.orangeColor())
        self.stringDestView.locationBox.show(False)

    def stringDestViewDropCandidateEndedCallback(self, info):
        self.stringDestView.box.setBorderWidth(0)
        self.stringDestView.locationBox.show(False)
        self.stringDestView.textBox.set("drop: string")

    def stringDestViewPerformDropCallback(self, info):
        print("dest1PerformDropCallback")
        return True

    # plist

    def plistDestViewDropCandidateEnteredCallback(self, info):
        self.plistDestView.box.setBorderColor(AppKit.NSColor.greenColor())
        self.plistDestView.box.setBorderWidth(20)
        self.plistDestView.locationBox.show(True)
        return self.dropOperation

    def plistDestViewDropCandidateCallback(self, info):
        sender = info["sender"]
        source = info["source"]
        items = info["items"]
        items = sender.getDropItemValues(
            items,
            "plist"
        )
        location = info["location"]
        x, y = location
        self.plistDestView.locationBox.setPosSize((x-10, y-10, 20, 20))
        text = (
            f"count={len(items)}",
            f"items={repr(items)}",
            f"source={source.__class__.__name__}"
        )
        text = "\n".join(text)
        self.plistDestView.textBox.set(text)
        return self.dropOperation

    def plistDestViewDropCandidateExitedCallback(self, info):
        self.plistDestView.box.setBorderColor(AppKit.NSColor.orangeColor())
        self.plistDestView.locationBox.show(False)

    def plistDestViewDropCandidateEndedCallback(self, info):
        self.plistDestView.box.setBorderWidth(0)
        self.plistDestView.locationBox.show(False)
        self.plistDestView.textBox.set("drop: string")

    def plistDestViewPerformDropCallback(self, info):
        print("dest2PerformDropCallback")
        return True

    # file urls

    def fileURLDestViewDropCandidateEnteredCallback(self, info):
        self.fileURLDestView.box.setBorderColor(AppKit.NSColor.greenColor())
        self.fileURLDestView.box.setBorderWidth(20)
        self.fileURLDestView.locationBox.show(True)
        return self.dropOperation

    def fileURLDestViewDropCandidateCallback(self, info):
        sender = info["sender"]
        source = info["source"]
        items = info["items"]
        items = sender.getDropItemValues(
            items,
            "fileURL"
        )
        location = info["location"]
        x, y = location
        self.fileURLDestView.locationBox.setPosSize((x-10, y-10, 20, 20))
        text = (
            f"count={len(items)}",
            f"items={repr(items)}",
            f"source={source.__class__.__name__}"
        )
        text = "\n".join(text)
        self.fileURLDestView.textBox.set(text)
        return self.dropOperation

    def fileURLDestViewDropCandidateExitedCallback(self, info):
        self.fileURLDestView.box.setBorderColor(AppKit.NSColor.orangeColor())
        self.fileURLDestView.locationBox.show(False)

    def fileURLDestViewDropCandidateEndedCallback(self, info):
        self.fileURLDestView.box.setBorderWidth(0)
        self.fileURLDestView.locationBox.show(False)
        self.fileURLDestView.textBox.set("drop: files")

    def fileURLDestViewPerformDropCallback(self, info):
        print("dest3PerformDropCallback")
        sender = info["sender"]
        items = info["items"]
        items = sender.getDropItemValues(
            items,
            "fileURL"
        )
        for item in items:
            print(item.path())
        return True

    # Destination Lists
    # -----------------

    def makeDestLists(self):
        dragSettings = dict(
            makeDragDataCallback=self.stringDestListMakeDragDataCallback
        )
        dropSettings = dict(
            pasteboardTypes=[
                "string",
                "dev.robotools.test.stringDestListIndexes"
            ],
            dropCandidateEnteredCallback=self.stringDestListDropCandidateEnteredCallback,
            dropCandidateEndedCallback=self.stringDestListDropCandidateEndedCallback,
            dropCandidateExitedCallback=self.stringDestListDropCandidateExitedCallback,
            dropCandidateCallback=self.stringDestListDropCandidateCallback,
            performDropCallback=self.stringDestListPerformDropCallback
        )
        self.stringDestList = vanilla.List2(
            "auto",
            ["XXX", "YYY", "ZZZ"],
            dragSettings=dragSettings,
            dropSettings=dropSettings
        )

        columnDescriptions = [
            dict(
                identifier="letter",
                title="ABC",
                width=35
            ),
            dict(
                identifier="number",
                title="#",
                width=35
            )
        ]
        dragSettings = dict(
            makeDragDataCallback=self.plistDestListMakeDragDataCallback
        )
        dropSettings = dict(
            pasteboardTypes=["plist"],
            dropCandidateEnteredCallback=self.plistDestListDropCandidateEnteredCallback,
            dropCandidateEndedCallback=self.plistDestListDropCandidateEndedCallback,
            dropCandidateExitedCallback=self.plistDestListDropCandidateExitedCallback,
            dropCandidateCallback=self.plistDestListDropCandidateCallback,
            performDropCallback=self.plistDestListPerformDropCallback
        )
        self.plistDestList = vanilla.List2(
            "auto",
            [
                dict(
                    letter="XXX",
                    number=-1
                ),
                dict(
                    letter="YYY",
                    number=-2
                ),
                dict(
                    letter="ZZZ",
                    number=-3
                ),
            ],
            columnDescriptions=columnDescriptions,
            dragSettings=dragSettings,
            dropSettings=dropSettings
        )

        dropSettings = dict(
            pasteboardTypes=["fileURL"],
            dropCandidateCallback=self.fileURLDestListDropCandidateCallback,
            performDropCallback=self.fileURLDestListPerformDropCallback
        )
        self.fileURLDestList = vanilla.List2(
            "auto",
            [],
            dropSettings=dropSettings
        )

        self.destViewListStack = vanilla.HorizontalStackView(
            "auto",
            views=[
                self.stringDestList,
                self.plistDestList,
                dict(
                    view=vanilla.VerticalLine("auto"),
                    height="fill",
                    width=1
                ),
                self.fileURLDestList
            ],
            spacing=20
        )

    # string

    def stringDestListDropCandidateEnteredCallback(self, info):
        self.stringDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.greenColor()
        )
        return "generic"

    def stringDestListDropCandidateExitedCallback(self, info):
        self.stringDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.orangeColor()
        )

    def stringDestListDropCandidateEndedCallback(self, info):
        self.stringDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.whiteColor()
        )

    def stringDestListMakeDragDataCallback(self, index):
        typesAndValues = {
            "string" : self.stringDestList.get()[index],
            "dev.robotools.test.stringDestListIndexes" : index
        }
        return typesAndValues

    def stringDestListDropCandidateCallback(self, info):
        source = info["source"]
        if source == self.stringDestList:
            return "move"
        return "copy"

    def stringDestListPerformDropCallback(self, info):
        sender = info["sender"]
        source = info["source"]
        index = info["index"]
        items = info["items"]
        # reorder
        if source == self.stringDestList:
            indexes = sender.getDropItemValues(items, "dev.robotools.test.stringDestListIndexes")
            # XXX this is wrong
            items = list(self.stringDestList.get())
            move = [items.pop(i) for i in indexes]
            # insert
            if index is not None:
                items[index:index] = move
            # append
            else:
                items += move
        else:
            items = sender.getDropItemValues(items, "string")
            allItems = list(self.stringDestList.get())
            # insert
            if index is not None:
                items = allItems[:index] + items + allItems[index:]
            # append
            else:
                items = allItems + items
        self.stringDestList.set(items)
        return True

    # plist

    def plistDestListDropCandidateEnteredCallback(self, info):
        self.plistDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.greenColor()
        )
        return "generic"

    def plistDestListDropCandidateExitedCallback(self, info):
        self.plistDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.orangeColor()
        )

    def plistDestListDropCandidateEndedCallback(self, info):
        self.plistDestList.getNSScrollView().setBackgroundColor_(
            AppKit.NSColor.whiteColor()
        )

    def plistDestListMakeDragDataCallback(self, index):
        typesAndValues = {
            "plist" : self.plistDestList.get()[index]
        }
        return typesAndValues

    def plistDestListDropCandidateCallback(self, info):
        source = info["source"]
        if source == self.plistDestList:
            return "none"
        return "copy"

    def plistDestListPerformDropCallback(self, info):
        sender = info["sender"]
        index = info["index"]
        items = info["items"]
        items = sender.getDropItemValues(items)
        allItems = list(self.plistDestList.get())
        items = allItems[:index] + items + allItems[index:]
        self.plistDestList.set(items)
        return True

    # file urls

    def fileURLDestListDropCandidateCallback(self, info):
        existing = list(self.fileURLDestList.get())
        sender = info["sender"]
        items = info["items"]
        items = sender.getDropItemValues(items)
        items = [os.path.basename(item.path()) for item in items]
        items = [item for item in items if item not in existing]
        if not items:
            return "none"
        return "link"

    def fileURLDestListPerformDropCallback(self, info):
        existing = list(self.fileURLDestList.get())
        sender = info["sender"]
        index = info["index"]
        items = info["items"]
        items = sender.getDropItemValues(items)
        items = [os.path.basename(item.path()) for item in items]
        items = [item for item in items if item not in existing]
        if index is not None:
            items = existing[:index] + items + existing[index:]
        self.fileURLDestList.set(items)
        return True


if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)



import pprint
import AppKit
import vanilla

class Test:

    def __init__(self):
        self.w = vanilla.Window((1000, 500))
        self.w.line = vanilla.VerticalLine("auto")

        items = []
        for i in range(3):
            d = dict(
                textField="ABC",
                editableTextField="XYZ",
                slider=25,
                editableSlider=75,
                checkBox=True,
                editableCheckBox=False,
                popUpButton=1,
                editablePopUpButton=2,
                image=AppKit.NSImage.imageNamed_(AppKit.NSImageNameTrashFull),
                segmentedButton=1,
                editableSegmentedButton=2,
                colorWell=AppKit.NSColor.redColor(),
                editableColorWell=AppKit.NSColor.greenColor(),
                custom=dict(
                    c=AppKit.NSColor.yellowColor(),
                    pub=1,
                    cb=False,
                    s=0.5
                )
            )
            items.append(d)

        columnDescriptions = [
            dict(
                identifier="textField",
                title="TextField",
                width=75,
                editable=False
            ),
            dict(
                identifier="editableTextField",
                title="TextField-E",
                width=75,
                editable=True
            ),
            dict(
                identifier="slider",
                title="Slider",
                width=100,
                cellClass=vanilla.SliderList2Cell,
                cellClassArguments=dict(
                    minValue=0,
                    maxValue=100
                ),
                editable=False,
            ),
            dict(
                identifier="editableSlider",
                title="Slider-E",
                width=100,
                cellClass=vanilla.SliderList2Cell,
                cellClassArguments=dict(
                    minValue=0,
                    maxValue=100
                ),
                editable=True,
            ),
            dict(
                identifier="checkBox",
                title="CheckBox",
                width=20,
                cellClass=vanilla.CheckBoxList2Cell,
                editable=False,
            ),
            dict(
                identifier="editableCheckBox",
                title="CheckBox-E",
                width=20,
                cellClass=vanilla.CheckBoxList2Cell,
                editable=True,
            ),
            dict(
                identifier="popUpButton",
                title="PopUpButton",
                width=100,
                cellClass=vanilla.PopUpButtonList2Cell,
                cellClassArguments=dict(
                    items=list("ABCDE")
                ),
                editable=False,
            ),
            dict(
                identifier="editablePopUpButton",
                title="PopUpButton-E",
                width=100,
                cellClass=vanilla.PopUpButtonList2Cell,
                cellClassArguments=dict(
                    items=list("ABCDE")
                ),
                editable=True,
            ),
            dict(
                identifier="image",
                title="Image",
                width=35,
                cellClass=vanilla.ImageList2Cell
            ),
            dict(
                identifier="segmentedButton",
                title="SegmentedButton",
                width=100,
                cellClass=vanilla.SegmentedButtonList2Cell,
                cellClassArguments=dict(
                    segmentDescriptions=[
                        dict(title="A"),
                        dict(title="B"),
                        dict(title="C")
                    ]
                ),
                editable=False,
            ),
            dict(
                identifier="editableSegmentedButton",
                title="SegmentedButton-E",
                width=100,
                cellClass=vanilla.SegmentedButtonList2Cell,
                cellClassArguments=dict(
                    segmentDescriptions=[
                        dict(title="A"),
                        dict(title="B"),
                        dict(title="C")
                    ]
                ),
                editable=True,
            ),
            dict(
                identifier="colorWell",
                title="ColorWell",
                width=50,
                cellClass=vanilla.ColorWellList2Cell,
                editable=False,
            ),
            dict(
                identifier="editableColorWell",
                title="ColorWell-E",
                width=50,
                cellClass=vanilla.ColorWellList2Cell,
                editable=True,
            ),
            dict(
                identifier="custom",
                title="Custom",
                width=100,
                cellClass=CustomList2Cell,
                editable=True,
            ),
        ]
        self.w.list = vanilla.List2(
            "auto",
            items=items,
            columnDescriptions=columnDescriptions,
            editCallback=self.editCallback
        )
        self.w.getButton = vanilla.Button(
            "auto",
            "Get Values",
            callback=self.getButtonCallback
        )
        rules = [
            "H:|-[list]-|",
            "H:[getButton]-|",
            "V:|-[list]-[getButton]-|"
        ]
        self.w.addAutoPosSizeRules(rules)
        self.w.open()

    def editCallback(self, sender):
        print("editCallback")

    def getButtonCallback(self, sender):
        for i, item in enumerate(self.w.list.get()):
            print(f"{i} ----------")
            for k, v in item.items():
                if k == "image":
                    continue
                v = pprint.pformat(v)
                print(f"{k} : {v}")
            print("\n")


class CustomList2Cell(vanilla.Box):

    def __init__(self, editable=True, callback=None):
        self.callback = callback
        super().__init__(
            "auto",
            cornerRadius=10
        )
        self.popUpButton = vanilla.PopUpButton(
            "auto",
            items=list("ABC"),
            callback=self.changedCallback
        )
        self.checkBox = vanilla.CheckBox(
            "auto",
            title="Test",
            callback=self.changedCallback
        )
        self.slider = vanilla.Slider(
            "auto",
            minValue=0,
            maxValue=1.0,
            callback=self.changedCallback
        )
        rules = [
            "H:|-space-[popUpButton]-space-|",
            "H:|-space-[checkBox]-space-|",
            "H:|-space-[slider]-space-|",
            "V:|-space-[popUpButton]-space-[checkBox]-space-[slider]-space-|",
        ]
        metrics = dict(
            space=10
        )
        self.addAutoPosSizeRules(rules, metrics)

    def changedCallback(self, sender):
        self.callback(self)

    def get(self):
        value = dict(
            c=self.getNSBox().backgroundColor(),
            pub=self.popUpButton.get(),
            cb=self.checkBox.get(),
            s=self.slider.get()
        )
        return value

    def set(self, value):
        self.getNSBox().setBackgroundColor_(value["c"])
        self.popUpButton.set(value["pub"])
        self.checkBox.set(value["cb"])
        self.slider.set(value["s"])


if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)



import objc
from AppKit import NSBox, NSColor, NSFont, NSSmallControlSize, NSNoTitle, NSLineBorder, NSBoxSeparator, NSBoxCustom
from vanilla.vanillaBase import VanillaBaseObject, _breakCycles, osVersionCurrent, osVersion10_10


class Box(VanillaBaseObject):

    """
    A bordered container for other controls.

    .. image:: /_images/Box.png

    To add a control to a box, simply set it as an attribute of the box.

    ::

        from vanilla import Window, Box, TextBox

        class BoxDemo:

            def __init__(self):
                self.w = Window((150, 70))
                self.w.box = Box((10, 10, -10, -10))
                self.w.box.text = TextBox((10, 10, -10, -10), "This is a box")
                self.w.open()

        BoxDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the box.

    **title** The title to be displayed above the box. Pass *None* if no title is desired.
    """

    allFrameAdjustments = {
        # Box does not have sizeStyle, but the
        # adjustment is differeent based on the
        # presence of a title.
        "Box-Titled": (-3, -4, 6, 4),
        "Box-None": (-3, -4, 6, 6)
    }

    nsBoxClass = NSBox

    def __init__(self, posSize, title=None,
            fillColor=None, borderColor=None, borderWidth=None,
            cornerRadius=None, margins=None
        ):
        self._setupView(self.nsBoxClass, posSize)
        if title:
            self._nsObject.setTitle_(title)
            if osVersionCurrent < osVersion10_10:
                self._nsObject.titleCell().setTextColor_(NSColor.blackColor())
            font = NSFont.systemFontOfSize_(NSFont.systemFontSizeForControlSize_(NSSmallControlSize))
            self._nsObject.setTitleFont_(font)
        else:
            self._nsObject.setTitlePosition_(NSNoTitle)
        if fillColor is not None:
            self.setFillColor(fillColor)
        if borderColor is not None:
            self.setBorderColor(borderColor)
        if borderWidth is not None:
            self.setBorderWidth(borderWidth)
        if cornerRadius is not None:
            self.setCornerRadius(cornerRadius)
        if margins is not None:
            self.setMargins(margins)

    def getNSBox(self):
        """
        Return the `NSBox`_ that this object wraps.

        .. _NSBox: https://developer.apple.com/documentation/appkit/nsbox?language=objc
        """
        return self._nsObject

    def _adjustPosSize(self, frame):
        # skip subclasses
        if self.__class__.__name__ == "Box":
            pos = self._nsObject.titlePosition()
            if pos != NSNoTitle:
                title = "Titled"
            else:
                title = "None"
            boxType = "Box-" + title
            self.frameAdjustments = self.allFrameAdjustments[boxType]
        return super(Box, self)._adjustPosSize(frame)

    def _getContentView(self):
        return self._nsObject.contentView()

    def _breakCycles(self):
        super(Box, self)._breakCycles()
        view = self._nsObject.contentView()
        if view is not None:
            _breakCycles(view)

    def setTitle(self, title):
        """
        Set the title of the box.
        """
        self._nsObject.setTitle_(title)

    def getTitle(self):
        """
        Get the title of the box.
        """
        return self._nsObject.title()

    def setFillColor(self, color):
        """
        Set the fill color of the box.
        """
        self._nsObject.setBoxType_(NSBoxCustom)
        self._nsObject.setFillColor_(color)

    def setBorderColor(self, color):
        """
        Set the border color of the box.
        """
        self._nsObject.setBoxType_(NSBoxCustom)
        self._nsObject.setBorderColor_(color)

    def setBorderWidth(self, value):
        """
        Set the border width of the box.
        """
        self._nsObject.setBoxType_(NSBoxCustom)
        self._nsObject.setBorderWidth_(value)

    def setCornerRadius(self, value):
        """
        Set the corner radius of the box.
        """
        self._nsObject.setBoxType_(NSBoxCustom)
        self._nsObject.setCornerRadius_(value)

    def setMargins(self, value):
        """
        Set the x, y margins for the content.
        """
        if isinstance(value, int):
            value = (value, value)
        self._nsObject.setContentViewMargins_(value)
        self._nsObject.sizeToFit()

class _Line(Box):

    nsBoxClass = NSBox

    def __init__(self, posSize):
        self._setupView(self.nsBoxClass, posSize)
        self._nsObject.setBorderType_(NSLineBorder)
        self._nsObject.setBoxType_(NSBoxSeparator)
        self._nsObject.setTitlePosition_(NSNoTitle)

    def _getContentView(self):
        return self._nsObject


class HorizontalLine(_Line):

    """
    A horizontal line.

    .. image:: /_images/HorizontalLine.png

    ::

        from vanilla import Window, HorizontalLine

        class HorizontalLineDemo:

            def __init__(self):
                self.w = Window((100, 20))
                self.w.line = HorizontalLine((10, 10, -10, 1))
                self.w.open()

        HorizontalLineDemo()

    **posSize** Tuple of form *(left, top, width, height)* representing
    the position and size of the line.

    +-------------------------+
    | **Standard Dimensions** |
    +===+=====================+
    | H | 1                   |
    +---+---------------------+
    """

    def __init__(self, posSize):
        super(HorizontalLine, self).__init__(posSize)


class VerticalLine(_Line):

    """
    A vertical line.

    .. image:: /_images/VerticalLine.png

    ::

        from vanilla import Window, VerticalLine

        class VerticalLineDemo:

            def __init__(self):
                self.w = Window((80, 100))
                self.w.line = VerticalLine((40, 10, 1, -10))
                self.w.open()

        VerticalLineDemo()

    **posSize** Tuple of form *(left, top, width, height)* representing
    the position and size of the line.

    +-------------------------+
    | **Standard Dimensions** |
    +===+=====================+
    | V | 1                   |
    +---+---------------------+
    """

    def __init__(self, posSize):
        super(VerticalLine, self).__init__(posSize)



from objc import selector
from objc import python_method
from Foundation import NSObject
from AppKit import NSAlert, NSSavePanel, NSOpenPanel, NSAlertStyleCritical, NSAlertStyleInformational, NSAlertStyleWarning, NSAlertFirstButtonReturn, NSAlertSecondButtonReturn, NSAlertThirdButtonReturn, NSOKButton, NSURL, NSImage


__all__ = ["message", "askYesNoCancel", "askYesNo", "getFile", "getFolder", "getFileOrFolder", "putFile"]


alertStyleMap = {
    None : NSAlertStyleInformational,
    "informational" : NSAlertStyleInformational,
    "critical" : NSAlertStyleCritical,
    # "warning" : NSAlertStyleWarning,  # no difference with
    # backwards compatible keys
    NSAlertStyleInformational : NSAlertStyleInformational,
    NSAlertStyleCritical : NSAlertStyleCritical,
    NSAlertStyleWarning : NSAlertStyleWarning
}


class BasePanel(NSObject):

    def initWithWindow_resultCallback_(cls, parentWindow=None, resultCallback=None):
        self = cls.init()
        self.retain()
        self._parentWindow = parentWindow
        self._resultCallback = resultCallback
        return self

    def windowWillClose_(self, notification):
        self.autorelease()


class BaseMessageDialog(BasePanel):

    def initWithWindow_resultCallback_(cls, parentWindow=None, resultCallback=None):
        self = super(BaseMessageDialog, cls).initWithWindow_resultCallback_(parentWindow, resultCallback)
        self.messageText = ""
        self.informativeText = ""
        self.alertStyle = NSAlertStyleInformational
        self.buttonTitlesValues = []
        self.accessoryView = None
        self.icon = None
        self.showsHelpCallback = None
        return self

    def initWithMessageText_informativeText_alertStyle_buttonTitlesValues_window_resultCallback_(self,
            messageText="", informativeText="", alertStyle=NSAlertStyleInformational, buttonTitlesValues=[], parentWindow=None, resultCallback=None):
        # make it backwards compatible
        import warnings
        warnings.warn(
            "'BaseMessageDiaglog.alloc().initWithMessageText_informativeText_alertStyle_buttonTitlesValues_window_resultCallback_' has been deprecated and will be removed."
            "Please update your code.",
            DeprecationWarning
        )
        self = self.initWithWindow_resultCallback_(parentWindow, resultCallback)
        self.messageText = messageText
        self.informativeText = informativeText
        self.alertStyle = alertStyle
        self.buttonTitlesValues = _mapButtonTitles(buttonTitlesValues)
        self.run()
        return self

    def run(self):
        self.alert = NSAlert.alloc().init()
        self.alert.setDelegate_(self)
        self.alert.setMessageText_(self.messageText)
        self.alert.setInformativeText_(self.informativeText)
        self.alert.setAlertStyle_(self.alertStyle)

        if self.accessoryView:
            self.alert.setAccessoryView_(self.accessoryView)
        if self.icon:
            self.alert.setIcon_(self.icon)
        if self.showsHelpCallback:
            self.alert.setShowsHelp_(True)

        for buttonTitle in self.buttonTitlesValues:
            self.alert.addButtonWithTitle_(buttonTitle["title"])
        self._value = None
        if self._parentWindow is None:
            code = self.alert.runModal()
            self._translateValue(code)
            if self._resultCallback is not None:
                self._resultCallback(self._value)
        else:
            self.alert.beginSheetModalForWindow_completionHandler_(self._parentWindow, self.completionHandler_)

    def completionHandler_(self, returnCode):
        self.alert.window().close()
        self._translateValue(returnCode)
        if self._resultCallback is not None:
            self._resultCallback(self._value)

    @python_method
    def _translateValue(self, code):
        if code == NSAlertFirstButtonReturn:
            value = 1
        elif code == NSAlertSecondButtonReturn:
            value = 2
        elif code == NSAlertThirdButtonReturn:
            value = 3
        else:
            value = code - NSAlertThirdButtonReturn + 3
        if self.buttonTitlesValues:
            result = self.buttonTitlesValues[value - 1]
            if "callback" in result:
                result["callback"]()
            self._value = result.get("returnCode")

    # delegate method

    def alertShowHelp_(self, sender):
        self.showsHelpCallback()


class BasePutGetPanel(BasePanel):

    def completionHandler_(self, returnCode):
        self.panel.close()
        if returnCode:
            self._result = self.panel.filenames()
            if self._resultCallback is not None:
                self._resultCallback(self._result)


class PutFilePanel(BasePutGetPanel):

    def initWithWindow_resultCallback_(self, parentWindow=None, resultCallback=None):
        self = super(PutFilePanel, self).initWithWindow_resultCallback_(parentWindow, resultCallback)
        self.messageText = None
        self.title = None
        self.fileTypes = None
        self.directory = None
        self.fileName = None
        self.canCreateDirectories = True
        self.accessoryView = None
        self._result = None
        return self

    def run(self):
        self.panel = NSSavePanel.alloc().init()
        if self.messageText:
            self.panel.setMessage_(self.messageText)
        if self.title:
            self.panel.setTitle_(self.title)
        if self.fileName:
            self.panel.setNameFieldStringValue_(self.fileName)
        if self.directory:
            self.panel.setDirectoryURL_(NSURL.fileURLWithPath_(self.directory))
        if self.fileTypes:
            self.panel.setAllowedFileTypes_(self.fileTypes)
        self.panel.setCanCreateDirectories_(self.canCreateDirectories)
        self.panel.setCanSelectHiddenExtension_(True)
        self.panel.setAccessoryView_(self.accessoryView)
        if self._parentWindow is not None:
            self.panel.beginSheetModalForWindow_completionHandler_(self._parentWindow, self.completionHandler_)
        else:
            isOK = self.panel.runModalForDirectory_file_(self.directory, self.fileName)
            if isOK == NSOKButton:
                self._result = self.panel.filename()

    def completionHandler_(self, returnCode):
        self.panel.close()
        if returnCode:
            self._result = self.panel.filename()
            if self._resultCallback is not None:
                self._resultCallback(self._result)


class GetFileOrFolderPanel(BasePutGetPanel):

    def initWithWindow_resultCallback_(self, parentWindow=None, resultCallback=None):
        self = super(GetFileOrFolderPanel, self).initWithWindow_resultCallback_(parentWindow, resultCallback)
        self.messageText = None
        self.title = None
        self.directory = None
        self.fileName = None
        self.fileTypes = None
        self.allowsMultipleSelection = False
        self.canChooseDirectories = True
        self.canChooseFiles = True
        self.resolvesAliases = True
        self.accessoryView = None
        self._result = None
        return self

    def run(self):
        self.panel = NSOpenPanel.alloc().init()
        if self.messageText:
            self.panel.setMessage_(self.messageText)
        if self.title:
            self.panel.setTitle_(self.title)
        if self.fileName:
            self.panel.setNameFieldLabel_(self.fileName)
        if self.directory:
            self.panel.setDirectoryURL_(NSURL.fileURLWithPath_(self.directory))
        if self.fileTypes:
            self.panel.setAllowedFileTypes_(self.fileTypes)
        self.panel.setCanChooseDirectories_(self.canChooseDirectories)
        self.panel.setCanChooseFiles_(self.canChooseFiles)
        self.panel.setAllowsMultipleSelection_(self.allowsMultipleSelection)
        self.panel.setResolvesAliases_(self.resolvesAliases)
        self.panel.setAccessoryView_(self.accessoryView)
        if self._parentWindow is not None:
            self.panel.beginSheetModalForWindow_completionHandler_(self._parentWindow, self.completionHandler_)
        else:
            isOK = self.panel.runModalForDirectory_file_types_(self.directory, self.fileName, self.fileTypes)
            if isOK == NSOKButton:
                self._result = self.panel.filenames()


def _unwrapWindow(window):
    from vanilla.vanillaWindows import Window
    if window is None:
        return window
    if isinstance(window, Window):
        window = window.getNSWindow()
    return window


def _unwrapView(view):
    from vanilla.vanillaBase import VanillaBaseObject
    if view is None:
        return view
    if isinstance(view, VanillaBaseObject):
        l, t, w, h = view.getPosSize()
        view = view._getContentView()
        view.setFrame_(((0, 0), (w, h)))
    return view


def _mapButtonTitles(titles):
    """
    Convert key
    """
    buttonTitles = []
    for buttonTitle in titles:
        if isinstance(buttonTitle, tuple):
            title, returnCode = buttonTitle
            buttonTitle = dict(title=title, returnCode=returnCode)
        buttonTitles.append(buttonTitle)
    return buttonTitles


def message(messageText="", informativeText="", alertStyle="informational", parentWindow=None, resultCallback=None, icon=None, accessoryView=None, showsHelpCallback=None):
    assert icon is None or isinstance(icon, NSImage)

    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)

    alert = BaseMessageDialog.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    alert.messageText = messageText
    alert.informativeText = informativeText
    alert.alertStyle = alertStyleMap[alertStyle]
    alert.accessoryView = accessoryView
    alert.icon = icon
    alert.showsHelpCallback = showsHelpCallback
    alert.run()

    if resultCallback is None:
        return 1


def ask(messageText="", informativeText="", alertStyle="informational", buttonTitles=[],
        parentWindow=None, resultCallback=None, icon=None, accessoryView=None, showsHelpCallback=None):
    assert buttonTitles, "Button titles are required"
    assert icon is None or isinstance(icon, NSImage)

    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)
    buttonTitles = _mapButtonTitles(buttonTitles)

    alert = BaseMessageDialog.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    alert.messageText = messageText
    alert.informativeText = informativeText
    alert.alertStyle = alertStyleMap[alertStyle]
    alert.buttonTitlesValues = buttonTitles
    alert.accessoryView = accessoryView
    alert.icon = icon
    alert.showsHelpCallback = showsHelpCallback
    alert.run()

    if resultCallback is None:
        return alert._value


def askYesNoCancel(messageText="", informativeText="", alertStyle="informational",
        parentWindow=None, resultCallback=None, icon=None, accessoryView=None, showsHelpCallback=None):
    return ask(
        messageText=messageText,
        informativeText=informativeText,
        alertStyle=alertStyle,
        buttonTitles=[
            dict(title="Cancel", returnCode=-1),
            dict(title="Yes", returnCode=1),
            dict(title="No", returnCode=0)
        ],
        parentWindow=parentWindow,
        resultCallback=resultCallback,
        icon=icon,
        accessoryView=accessoryView,
        showsHelpCallback=showsHelpCallback
    )


def askYesNo(messageText="", informativeText="", alertStyle="informational",
        parentWindow=None, resultCallback=None, icon=None, accessoryView=None, showsHelpCallback=None):
    return ask(
        messageText=messageText,
        informativeText=informativeText,
        alertStyle=alertStyle,
        buttonTitles=[
            dict(title="Yes", returnCode=1),
            dict(title="No", returnCode=0)
        ],
        parentWindow=parentWindow,
        resultCallback=resultCallback,
        icon=icon,
        accessoryView=accessoryView,
        showsHelpCallback=showsHelpCallback
    )


def getFile(messageText=None, title=None, directory=None, fileName=None,
        allowsMultipleSelection=False, fileTypes=None, parentWindow=None, resultCallback=None, accessoryView=None):
    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)

    basePanel = GetFileOrFolderPanel.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    basePanel.messageText = messageText
    basePanel.title = title
    basePanel.directory = directory
    basePanel.fileName = fileName
    basePanel.fileTypes = fileTypes
    basePanel.allowsMultipleSelection = allowsMultipleSelection
    basePanel.canChooseDirectories = False
    basePanel.canChooseFiles = True
    basePanel.accessoryView = accessoryView
    basePanel.run()

    if resultCallback is None:
        return basePanel._result


def getFolder(messageText=None, title=None, directory=None, allowsMultipleSelection=False,
        parentWindow=None, resultCallback=None, accessoryView=None):
    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)

    basePanel = GetFileOrFolderPanel.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    basePanel.messageText = messageText
    basePanel.title = title
    basePanel.directory = directory
    basePanel.allowsMultipleSelection = allowsMultipleSelection
    basePanel.canChooseDirectories = True
    basePanel.canChooseFiles = False
    basePanel.accessoryView = accessoryView
    basePanel.run()

    if resultCallback is None:
        return basePanel._result


def getFileOrFolder(messageText=None, title=None, directory=None, fileName=None,
        allowsMultipleSelection=False, fileTypes=None, parentWindow=None, resultCallback=None, accessoryView=None):
    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)

    basePanel = GetFileOrFolderPanel.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    basePanel.messageText = messageText
    basePanel.title = title
    basePanel.directory = directory
    basePanel.fileName = fileName
    basePanel.fileTypes = fileTypes
    basePanel.allowsMultipleSelection = allowsMultipleSelection
    basePanel.canChooseDirectories = True
    basePanel.canChooseFiles = True
    basePanel.accessoryView = accessoryView
    basePanel.run()

    if resultCallback is None:
        return basePanel._result


def putFile(messageText=None, title=None, directory=None, fileName=None, canCreateDirectories=True,
        fileTypes=None, parentWindow=None, resultCallback=None, accessoryView=None):
    parentWindow = _unwrapWindow(parentWindow)
    accessoryView = _unwrapView(accessoryView)

    basePanel = PutFilePanel.alloc().initWithWindow_resultCallback_(parentWindow, resultCallback)
    basePanel.messageText = messageText
    basePanel.title = title
    basePanel.directory = directory
    basePanel.fileName = fileName
    basePanel.fileTypes = fileTypes
    basePanel.canCreateDirectories = canCreateDirectories
    basePanel.accessoryView = accessoryView
    basePanel.run()

    if resultCallback is None:
        return basePanel._result



from AppKit import NSSegmentedControl, NSSegmentedCell, NSImage, NSSegmentSwitchTrackingSelectOne, NSSegmentSwitchTrackingSelectAny, NSSegmentSwitchTrackingMomentary

from vanilla.vanillaBase import VanillaBaseControl


_trackingModeMap = {
    "one": NSSegmentSwitchTrackingSelectOne,
    "any": NSSegmentSwitchTrackingSelectAny,
    "momentary": NSSegmentSwitchTrackingMomentary,
}


class SegmentedButton(VanillaBaseControl):

    """
    A standard segmented button.

    .. image:: /_images/SegmentedButton.png

    ::

        from vanilla import Window, SegmentedButton

        class SegmentedButtonDemo:

             def __init__(self):
                 self.w = Window((120, 40))
                 self.w.button = SegmentedButton((10, 10, -10, 20),
                     [dict(title="A"), dict(title="B"), dict(title="C")],
                    callback=self.buttonCallback)
                 self.w.open()

             def buttonCallback(self, sender):
                 print("button hit!")

        SegmentedButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the segmented button. The size of the segmented button
    should match the appropriate value for the given *sizeStyle*.

    +-------------------------+
    | **Standard Dimensions** |
    +=========+===+===========+
    | Regular | H | 21        |
    +---------+---+-----------+
    | Small   | H | 18        |
    +---------+---+-----------+
    | Mini    | H | 15        |
    +---------+---+-----------+

    **segmentDescriptions** An ordered list of dictionaries describing the segments.

    +----------------------------+----------------------------------------------------------+
    | width (optional)           | The desired width of the segment.                        |
    +----------------------------+----------------------------------------------------------+
    | title (optional)           | The title of the segment.                                |
    +----------------------------+----------------------------------------------------------+
    | enabled (optional)         | The enabled state of the segment. The default is `True`. |
    +----------------------------+----------------------------------------------------------+
    | imagePath (optional)       | A file path to an image to display in the segment.       |
    +----------------------------+----------------------------------------------------------+
    | imageNamed (optional)      | The name of an image already loaded as a `NSImage`_ by   |
    |                            | the application to display in the segment.               |
    +----------------------------+----------------------------------------------------------+
    | imageObject (optional)     | A `NSImage`_ object to display in the segment.           |
    +----------------------------+----------------------------------------------------------+
    | *imageTemplate* (optional) | A boolean representing if the image should converted     |
    |                            | to a template image.                                     |
    +----------------------------+----------------------------------------------------------+

    **callback** The method to be called when the user presses the segmented button.

    **selectionStyle** The selection style in the segmented button.

    +-----------+---------------------------------------------+
    | one       | Only one segment may be selected.           |
    +-----------+---------------------------------------------+
    | any       | Any number of segments may be selected.     |
    +-----------+---------------------------------------------+
    | momentary | A segmented is only selected when tracking. |
    +-----------+---------------------------------------------+

    **sizeStyle** A string representing the desired size style of the segmented button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc
    """

    nsSegmentedControlClass = NSSegmentedControl
    nsSegmentedCellClass = NSSegmentedCell

    frameAdjustments = {
        "mini": (0, -1, 0, 1), #15
        "small": (-2, -4, 2, 5), #20
        "regular": (0, -4, 0, 5), #24
        }

    def __init__(self, posSize, segmentDescriptions, callback=None, selectionStyle="one", sizeStyle="small"):
        self._setupView(self.nsSegmentedControlClass, posSize)
        if self.nsSegmentedCellClass != NSSegmentedCell:
            self._nsObject.setCell_(self.nsSegmentedCellClass.alloc().init())
        if callback is not None:
            self._setCallback(callback)
        self._setSizeStyle(sizeStyle)
        nsObject = self._nsObject
        nsObject.setSegmentCount_(len(segmentDescriptions))
        nsObject.cell().setTrackingMode_(_trackingModeMap[selectionStyle])
        for segmentIndex, segmentDescription in enumerate(segmentDescriptions):
            width = segmentDescription.get("width", 0)
            title = segmentDescription.get("title", "")
            enabled = segmentDescription.get("enabled", True)
            imagePath = segmentDescription.get("imagePath")
            imageNamed = segmentDescription.get("imageNamed")
            imageTemplate = segmentDescription.get("imageTemplate")
            imageObject = segmentDescription.get("imageObject")
            # create the NSImage if needed
            if imagePath is not None:
                image = NSImage.alloc().initWithContentsOfFile_(imagePath)
            elif imageNamed is not None:
                image = NSImage.imageNamed_(imageNamed)
            elif imageObject is not None:
                image = imageObject
            else:
                image = None
            nsObject.setWidth_forSegment_(width, segmentIndex)
            nsObject.setLabel_forSegment_(title, segmentIndex)
            nsObject.setEnabled_forSegment_(enabled, segmentIndex)
            if image is not None:
                if imageTemplate is not None:
                    # only change the image template setting if its either True or False
                    image.setTemplate_(imageTemplate)
                nsObject.setImage_forSegment_(image, segmentIndex)

    def getNSSegmentedButton(self):
        """
        Return the `NSSegmentedControl`_ that this object wraps.

        .. _NSSegmentedControl: https://developer.apple.com/documentation/appkit/nssegmentedcontrol?language=objc
        """
        return self._nsObject

    def enable(self, onOff):
        """
        Enable or disable the object. **onOff** should be a boolean.
        """
        for index in range(self._nsObject.segmentCount()):
            self._nsObject.setEnabled_forSegment_(onOff, index)

    def set(self, value):
        """
        Set the selected segment. If this control is set to
        `any` mode, `value` should be a list of integers.
        Otherwise `value` should be a single integer.
        """
        # value should be an int unless we are in "any" mode
        if self._nsObject.cell().trackingMode() != _trackingModeMap["any"]:
            value = [value]
        for index in range(self._nsObject.segmentCount()):
            state = index in value
            self._nsObject.setSelected_forSegment_(state, index)

    def get(self):
        """
        Get the selected segment. If this control is set to
        `any` mode, the returned value will be a list of integers.
        Otherwise the returned value will be a single integer or
        `None` if no segment is selected.
        """
        states = []
        for index in range(self._nsObject.segmentCount()):
            state = self._nsObject.isSelectedForSegment_(index)
            if state:
                states.append(index)
        if self._nsObject.cell().trackingMode() != _trackingModeMap["any"]:
            if states:
                return states[0]
            return None
        return states




import objc
import AppKit
from vanilla.vanillaBase import VanillaBaseObject

# Dragging Source

draggingFormationMap = dict(
    default=AppKit.NSDraggingFormationDefault,
    none=AppKit.NSDraggingFormationNone,
    pile=AppKit.NSDraggingFormationPile,
    stack=AppKit.NSDraggingFormationStack,
    list=AppKit.NSDraggingFormationList
)
try:
    NSPasteboardTypeFileURL = AppKit.NSPasteboardTypeFileURL
except AttributeError:
    NSPasteboardTypeFileURL = "public.file-url"

pasteboardTypeMap = dict(
    string=AppKit.NSPasteboardTypeString,
    plist="dev.robotools.vanilla.propertyList",
    fileURL=NSPasteboardTypeFileURL
)

def startDraggingSession(
        view,
        event,
        items,
        source=None,
        formation="default",
        location=None
    ):
    """
    Start a dragging session.

    `view` The view that is initiating the session.
    This may be a vanilla object or an instance of NSView.

    `event` The mouse event that is initiating the session.

    `items` A list of item definition dictionaries.

    - `typesAndValues` A dictionary of pasteboard types
      as keys and the object value for the type as the value.
      Acceptable value types are strings, bytes and any python
      object structure that can be represented as a Property List.
    - `location` The location within the source view from
      which the dragging image should originate. Optional.
    - `size` The size to display the dragging image. Optional.
    - `image` An image previewing the item. Optional.

    `source` A source for the dragging data if it is something
    other than the view that initiated the session. Optional.

    `formation` The formation of the dragging images. Options:

    - "default"
    - "none"
    - "pile"
    - "stack"
    - "list"

    `location` A fallback in case `location` is not defined in
    an item dictionary. Optional.
    """
    if isinstance(view, VanillaBaseObject):
        view = view._nsObject
    if source is None:
        source = view
    draggingItems = []
    if location is None:
        rect = view.visibleRect()
        x = AppKit.NSMidX(rect)
        y = AppKit.NSMidY(rect)
        location = (x, y)
    for item in items:
        if "location" not in item:
            item["location"] = location
        draggingItem = _makeDraggingItem(**item)
        draggingItems.append(draggingItem)
    session = view.beginDraggingSessionWithItems_event_source_(
        draggingItems,
        event,
        source
    )
    formation = draggingFormationMap.get(formation, formation)
    session.setDraggingFormation_(formation)
    return session

def makePasteboardItem(typesAndValues):
    pasteboardItem = AppKit.NSPasteboardItem.alloc().init()
    for pasteboardType, value in typesAndValues.items():
        pasteboardType = pasteboardTypeMap.get(pasteboardType, pasteboardType)
        if isinstance(value, bytes):
            pasteboardItem.setData_forType_(value, pasteboardType)
        elif isinstance(value, str):
            pasteboardItem.setString_forType_(value, pasteboardType)
        else:
            pasteboardItem.setPropertyList_forType_(value, pasteboardType)
    return pasteboardItem

def _makeDraggingItem(
        typesAndValues,
        location=(0, 0),
        size=None,
        image=None
    ):
    pasteboardItem = makePasteboardItem(typesAndValues)
    if size is None:
        if image is not None:
            size = image.size()
        else:
            size = (20, 20)
    draggingItem = AppKit.NSDraggingItem.alloc().initWithPasteboardWriter_(pasteboardItem)
    draggingItem.setDraggingFrame_contents_(
        (location, size),
        image
    )
    return draggingItem


# Drop Target

dropOperationMap = dict(
    none=AppKit.NSDragOperationNone,
    copy=AppKit.NSDragOperationCopy,
    link=AppKit.NSDragOperationLink,
    generic=AppKit.NSDragOperationGeneric,
    private=AppKit.NSDragOperationPrivate,
    move=AppKit.NSDragOperationMove,
    delete=AppKit.NSDragOperationDelete,
    drag=AppKit.NSDragOperationEvery
)

class DropTargetProtocolMixIn:

    """
    # Drop Settings

    `pasteboardTypes`

    A list of `NSPasteboardType` values, convenience values
    or custom strings defining the pasteboard types this
    view will accept during drops.

    - "string"
    - "plist"
    - "fileURL"

    `dropCandidateCallback`

    Optional. A method that will be called when one of the
    other `dropCandidate` callbacks is not defined. This must
    return one of the drop operations listed below.

    `dropCandidateEnteredCallback`

    Optional. A method that will be called when a drag
    operation enters the view. This must return one of
    the drop operations listed below.

    `dropCandidateUpdatedCallback`

    Optional. A method that will be called when a drag
    operation updates within the view. This must return
    one of the drop operations listed below.

    `dropCandidateExitedCallback`

    Optional. A method that will be called when a drag
    operation exits the view.

    `dropCandidateEndedCallback`

    Optional. A method that will be called when a drag
    operation that previously entered the view ends.

    `prepareForDropCallback`

    Optional. A method that will be called at the start of
    a drop. This should not drop the incoming data. This
    must return a boolean indicating if the drop is acceptable.

    `performDropCallback`

    A method that will be called when the drop is performed.
    This must return a boolean indicating if the drop was accepted.

    `finishDropCallback`

    Optional. A method that will be called after the drop has
    been performed.

    ## Drop Operations:

    These define the drop operation that the view will
    perform if the drop candidate is actually dropped.

    - "none"
    - "copy"
    - "link"
    - "generic"
    - "private"
    - "move"
    - "delete"
    - "drag"

    ## Dragging Info

    The callbacks will be sent a single `draggingInfo` argument.
    The value of this will be a dict with these key/value pairs:

    - `sender` The vanilla wrapper of the view being targeted
      for the drop.
    - `source` The source view that initiated the dragging operation.
      If this view is wrapped by vanilla, the vanilla wrapper
      will be given as the source.
    - `items` The list of NSPasteboardItem objects. Use the
      `getDropItemValues` method to unpack these to Python objects.
    - `location` The location of the drop.
    - `draggingInfo` The underlying NSDraggingInfo object.
    """

    # Drop Setup

    _dropCandidateCallback = None
    _dropCandidateEnteredCallback = None
    _dropCandidateUpdatedCallback = None
    _dropCandidateEndedCallback = None
    _dropCandidateExitedCallback = None
    _prepareForDropCallback = None
    _performDropCallback = None
    _finishDropCallback = None

    def _getDropView(self):
        return self._nsObject

    def setDropSettings(self, settings):
        pasteboardTypes = settings.get("pasteboardTypes")
        performDropCallback = settings.get("performDropCallback")
        assert pasteboardTypes is not None
        assert performDropCallback is not None
        self._registerDropTypes(pasteboardTypes)
        self._dropCandidateCallback = settings.get("dropCandidateCallback")
        self._dropCandidateEnteredCallback = settings.get("dropCandidateEnteredCallback")
        self._dropCandidateUpdatedCallback = settings.get("dropCandidateUpdatedCallback")
        self._dropCandidateEndedCallback = settings.get("dropCandidateEndedCallback")
        self._dropCandidateExitedCallback = settings.get("dropCandidateExitedCallback")
        self._prepareForDropCallback = settings.get("prepareForDropCallback")
        self._performDropCallback = performDropCallback
        self._finishDropCallback = settings.get("finishDropCallback")

    def _registerDropTypes(self, pasteboardTypes):
        view = self._getDropView()
        unwrapped = []
        for pasteboardType in pasteboardTypes:
            pasteboardType = pasteboardTypeMap.get(pasteboardType, pasteboardType)
            unwrapped.append(pasteboardType)
        view.registerForDraggedTypes_(unwrapped)

    def getDropItemValues(self, items, pasteboardType=None):
        """
        Get Python objects from the given NSPasteboardItem
        objects for the given pasteboard type. If this view
        is registered for only one pasteboard type, None may
        be given as the pasteboard type.
        """
        if pasteboardType is None:
            pasteboardTypes = self._getDropView().registeredDraggedTypes()
            assert len(pasteboardTypes) == 1
            pasteboardType = pasteboardTypes[0]
        pasteboardType = pasteboardTypeMap.get(pasteboardType, pasteboardType)
        values = []
        for item in items:
            if pasteboardType == AppKit.NSPasteboardTypeFileURL:
                value = item.stringForType_(pasteboardType)
                value = AppKit.NSURL.URLWithString_(value)
            else:
                value = item.propertyListForType_(pasteboardType)
            values.append(value)
        return values

    # Drop Candidate

    def _dropCandidateCallbackCaller(self, draggingInfo, callback, draggingEvent):
        if callback is None:
            callback = self._dropCandidateCallback
        if callback is None:
            value = "none"
        else:
            info = self._unpackDropCandidateInfo(draggingInfo)
            info["draggingEvent"] = draggingEvent
            value = callback(info)
        return value

    def _unpackDropCandidateInfo(self, draggingInfo):
        source = draggingInfo.draggingSource()
        if hasattr(source, "vanillaWrapper"):
            source = source.vanillaWrapper()
        pasteboard = draggingInfo.draggingPasteboard()
        items = pasteboard.pasteboardItems()
        x, y = self._getDropView().convertPoint_fromView_(
            draggingInfo.draggingLocation(),
            None
        )
        w, h = self._getDropView().frame().size
        location = (x, h - y)
        info = dict(
            sender=self,
            source=source,
            items=items,
            location=location,
            draggingInfo=draggingInfo
        )
        return info

    def _dropCandidateEntered(self, draggingInfo):
        operation = self._dropCandidateCallbackCaller(
            draggingInfo,
            self._dropCandidateEnteredCallback,
            draggingEvent="entered"
        )
        return dropOperationMap.get(operation, operation)

    def _dropCandidateUpdated(self, draggingInfo):
        operation = self._dropCandidateCallbackCaller(
            draggingInfo,
            self._dropCandidateUpdatedCallback,
            draggingEvent="updated"
        )
        return dropOperationMap.get(operation, operation)

    def _dropCandidateEnded(self, draggingInfo):
        self._dropCandidateCallbackCaller(
            draggingInfo,
            self._dropCandidateEndedCallback,
            draggingEvent="ended"
        )

    def _dropCandidateExited(self, draggingInfo):
        self._dropCandidateCallbackCaller(
            draggingInfo,
            self._dropCandidateExitedCallback,
            draggingEvent="exited"
        )

    def _updateDropCandidateImages(self, draggingInfo):
        """
        XXX

        What I'd like to do in the callback:
        - have a method for getting (item, image) pairs
          for items on the pasteboard with a given type
        - return a list of (item, image [or None]) pairs
        Then use the enumerator to update the images.
        """
        # def imageEnumerator(draggingItem, index, stop):
        #     xxx
        #     print(index)
        #     pass
        # draggingInfo.enumerateDraggingItemsWithOptions_forView_classes_searchOptions_usingBlock_(
        #     AppKit.NSDraggingItemEnumerationConcurrent,
        #     self._getContentView(),
        #     None,
        #     None,
        #     imageEnumerator
        # )
        # print("updateDropCandidateImages")

    # Drop Operation

    def _dropOperationCallbackCaller(self, draggingInfo, callback, draggingEvent):
        if callback is None:
            callback = self._dropCandidateCallback
        if callback is None:
            value = "none"
        else:
            info = self._unpackDropCandidateInfo(draggingInfo)
            info["draggingEvent"] = draggingEvent
            value = callback(info)
        return value

    def _prepareForDrop(self, draggingInfo):
        if self._prepareForDropCallback is None:
            return True
        info = self._unpackDropCandidateInfo(draggingInfo)
        return self._prepareForDropCallback(info)

    def _performDrop(self, draggingInfo):
        info = self._unpackDropCandidateInfo(draggingInfo)
        return self._performDropCallback(info)

    def _finishDrop(self, draggingInfo):
        if self._finishDropCallback is None:
            return
        info = self._unpackDropCandidateInfo(draggingInfo)
        self._finishDropCallback(info)



import operator
import weakref
import types
import objc
from objc import python_method
import AppKit
from vanilla.nsSubclasses import getNSSubclass
from vanilla.vanillaBase import VanillaBaseObject, VanillaCallbackWrapper, osVersionCurrent, osVersion10_16
from vanilla.vanillaScrollView import ScrollView
from vanilla.dragAndDrop import DropTargetProtocolMixIn, dropOperationMap, makePasteboardItem

simpleDataTypes = (
    str,
    int,
    float,
    complex,
    bool,
    type(None)
)

class VanillaList2DataSourceAndDelegate(AppKit.NSObject):

    def initWithTableView_(self, tableView):
        self = VanillaList2DataSourceAndDelegate.alloc().init()
        self._items = []
        self._arrangedIndexes = []
        self._groupRowIndexes = []
        self._tableView = tableView
        self._cellClasses = {} # { identifier : (class, kwargs) }
        self._valueToCellConverters = {} # { identifier : function }
        self._cellToValueConverters = {} # { identifier : function }
        self._groupRowCellClass = None
        self._groupRowCellClassKwargs = {}
        self._cellWrappers = {} # { nsView : vanilla wrapper } for view + wrapper reuse purposes
        self._valueGetters = {} # { identifier : options (see below) }
        self._valueSetters = {} # { identifier : options (see below) }
        # {
        #     property : str
        #     method : str
        #     function : func
        # }
        return self

    @python_method
    def setCellClassWithKwargsForColumn(self, cls, kwargs, identifier):
        if "callback" in kwargs:
            kwargs["callback"] = self.cellEditCallback
        self._cellClasses[identifier] = (cls, kwargs)

    @python_method
    def setGroupCellClassWithKwargs(self, cls, kwargs):
        self._groupRowCellClass = cls
        self._groupRowCellClassKwargs = kwargs

    @python_method
    def setGetters(self, getters):
        self._valueGetters = getters

    @python_method
    def setSetters(self, setters):
        self._valueSetters = setters

    @python_method
    def setValueToCellConverters(self, converters):
        self._valueToCellConverters = converters

    @python_method
    def setCellToValueConverters(self, converters):
        self._cellToValueConverters = converters

    @python_method
    def vanillaWrapper(self):
        return self._tableView.vanillaWrapper()

    @python_method
    def items(self):
        return self._items

    @python_method
    def setItems(self, items):
        self._items = items
        self._groupRowIndexes = []
        for index, item in enumerate(items):
            if isinstance(item, List2GroupRow):
                self._groupRowIndexes.append(index)
        self._updateArrangedIndexes()

    @python_method
    def arrangedIndexes(self):
        return self._arrangedIndexes

    @python_method
    def arrangedItems(self):
        items = [
            self._items[index]
            for index in self._arrangedIndexes
        ]
        return items

    @python_method
    def _updateArrangedIndexes(self):
        tableView = self._tableView
        sortDescriptors = tableView.sortDescriptors()
        items = self._items
        indexes = range(len(items))
        for sortDescriptor in reversed(sortDescriptors):
            identifier = sortDescriptor.key()
            ascending = sortDescriptor.ascending()
            reverse = not ascending
            getters = self._valueGetters.get(identifier, {})
            property = getters.get("property")
            method = getters.get("method")
            function = getters.get("function")
            if property is not None:
                key = operator.attrgetter(property)
            elif method is not None:
                key = operator.methodcaller(method)
            elif function is not None:
                key = function
            else:
                key = operator.itemgetter(identifier)

            def valueGetter(index):
                return key(items[index])

            indexes = sorted(indexes, key=valueGetter, reverse=reverse)
        self._arrangedIndexes = list(indexes)
        tableView.reloadData()

    @python_method
    def _getItemForRow(self, index):
        itemIndex = self._arrangedIndexes[index]
        return self._items[itemIndex]

    @python_method
    def getGroupValueForRow(self, index):
        itemIndex = self._arrangedIndexes[index]
        return self._items[itemIndex].value

    # Data Editing Via Cells

    @python_method
    def getItemValueForColumnAndRow(self, identifier, row):
        item = self._getItemForRow(row)
        getters = self._valueGetters.get(identifier, {})
        property = getters.get("property")
        method = getters.get("method")
        function = getters.get("function")
        if property is not None:
            value = getattr(item, property)
        elif method is not None:
            value = getattr(item, method)()
        elif function is not None:
            value = function(item)
        else:
            value = item[identifier]
        if identifier in self._valueToCellConverters:
            value = self._valueToCellConverters[identifier](value)
        return value

    @python_method
    def setItemValueForColumnAndRow(self, value, identifier, row):
        if identifier in self._cellToValueConverters:
            value = self._cellToValueConverters[identifier](value)
        item = self._getItemForRow(row)
        setters = self._valueSetters.get(identifier, {})
        property = setters.get("property")
        method = setters.get("method")
        function = setters.get("function")
        if property is not None:
            return setattr(item, property, value)
        elif method is not None:
            return getattr(item, method)(value)
        elif function is not None:
            return function(item, value)
        elif isinstance(item, dict):
            item[identifier] = value

    # Data Source

    def numberOfRowsInTableView_(self, tableView):
        return len(self._arrangedIndexes)

    def tableView_objectValueForTableColumn_row_(self, tableView, column, row):
        isGroupRow = column is None
        if isGroupRow:
            return self.getGroupValueForRow(row)
        identifier = column.identifier()
        value = self.getItemValueForColumnAndRow(identifier, row)
        return value

    def tableView_sortDescriptorsDidChange_(self, tableView, sortDescriptors):
        self._updateArrangedIndexes()

    # Delegate

    def tableView_viewForTableColumn_row_(self, tableView, column, row):
        isGroupRow = column is None
        if isGroupRow:
            view = self._groupRowCellClass(**self._groupRowCellClassKwargs)
            view.set(self.getGroupValueForRow(row))
            return view._nsObject
        identifier = column.identifier()
        value = self.getItemValueForColumnAndRow(identifier, row)
        nsView = tableView.makeViewWithIdentifier_owner_(
            column.identifier(),
            self
        )
        if nsView is not None:
            view = self._cellWrappers[nsView]
        else:
            cellClass, kwargs = self._cellClasses[identifier]
            view = cellClass(**kwargs)
            nsView = view._nsObject
            self._cellWrappers[nsView] = view
            nsView.setIdentifier_(identifier)
        view._representedColumnRow = (identifier, row)
        view.set(value)
        return nsView

    def tableViewSelectionDidChange_(self, notification):
        wrapper = self.vanillaWrapper()
        if wrapper._selectionCallback is not None:
            wrapper._selectionCallback(wrapper)

    def tableView_isGroupRow_(self, tableView, row):
        itemIndex = self._arrangedIndexes[row]
        return itemIndex in self._groupRowIndexes

    def tableView_shouldSelectRow_(self, tableView, row):
        wrapper = self.vanillaWrapper()
        if not wrapper._allowsSelection:
            return False
        return True

    # Editing

    @python_method
    def cellEditCallback(self, sender):
        identifier, row = sender._representedColumnRow
        value = sender.get()
        self.setItemValueForColumnAndRow(value, identifier, row)
        wrapper = self.vanillaWrapper()
        if wrapper._editCallback is not None:
            wrapper._editCallback(wrapper)

    # Drag

    def tableView_pasteboardWriterForRow_(
            self,
            tableView,
            row
        ):
        index = self._arrangedIndexes[row]
        return self.vanillaWrapper()._getPasteboardDataForIndex(index)

    # Drop

    def tableView_validateDrop_proposedRow_proposedDropOperation_(
            self,
            tableView,
            draggingInfo,
            row,
            operation
        ):
        if not self._arrangedIndexes:
            index = None
        elif row == len(self._arrangedIndexes):
            index = row
        else:
            index = self._arrangedIndexes[row]
        return self.vanillaWrapper()._dropCandidateUpdated(draggingInfo, index, operation)

    def tableView_acceptDrop_row_dropOperation_(
            self,
            tableView,
            draggingInfo,
            row,
            operation
        ):
        if not self._arrangedIndexes:
            index = None
        elif row == len(self._arrangedIndexes):
            index = row
        else:
            index = self._arrangedIndexes[row]
        return self.vanillaWrapper()._performDrop(draggingInfo, index, operation)


class VanillaList2TableViewSubclass(AppKit.NSTableView):

    def draggingEntered_(self, draggingInfo):
        super().draggingEntered_(draggingInfo)
        return self.vanillaWrapper()._dropCandidateEntered(draggingInfo)

    @objc.signature(b"Z@:@") # PyObjC bug? <- Found in the FontGoogles source.
    def draggingEnded_(self, draggingInfo):
        self.vanillaWrapper()._dropCandidateEnded(draggingInfo)
        # XXX
        # super().draggingEnded_(draggingInfo)
        # the super doesn't have this.
        # i'm not sure if it because of the
        # PyObjC bug mentioned above or
        # something else.

    def draggingExited_(self, draggingInfo):
        self.vanillaWrapper()._dropCandidateExited(draggingInfo)
        super().draggingExited_(draggingInfo)

    def canDragRowsWithIndexes_atPoint_(self, indexes, point):
        return self.vanillaWrapper()._validateRowsForDrag(indexes)


class List2(ScrollView, DropTargetProtocolMixIn):

    """
    A control that shows a list of items. These lists can contain one or more columns.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the list.

    **items** The items to be displayed in the list. This can be a simple list of
    basic object types:

    - str
    - int
    - float
    - complex
    - bool
    - None

    It may also be a list of dictionaries or other Python objects. The values will
    be retrieved and set with the provided value and getter and setter defined in the
    column description. If no getter or setter is given, the column identifier will be
    used to get and set the value in the item as if the item were a dictionary::

        value = item[columnIdentifier]
        item[columnIdentifier] = value

    **columnDescriptions** An ordered list of dictionaries describing the
    columns. This is only necessary for multiple column lists.

    +-------------------------------------+-----------------------------------------------------------+
    | *"identifier"*                      | The unique identifier for this column.                    |
    +-------------------------------------+-----------------------------------------------------------+
    | *"title"* (optional)                | The title to appear in the column header.                 |
    +-------------------------------------+-----------------------------------------------------------+
    | *"cellClass"* (optional)            | A cell class to be displayed in the column.               |
    |                                     | If nothing is given, a text cell is used.                 |
    +-------------------------------------+-----------------------------------------------------------+
    | *"cellClassArguments"* (optional)   | A dictionary of keyword arguments to be used when         |
    |                                     | *cellClass* is instantiated.                              |
    +-------------------------------------+-----------------------------------------------------------+
    | *"valueToCellConverter"* (optional) | A function for converting the value for display in the    |
    |                                     | cell.                                                     |
    +-------------------------------------+-----------------------------------------------------------+
    | *"cellToValueConverter"* (optional) | A function for converting the value displayed in the cell |
    |                                     | value for storage.                                        |
    +-------------------------------------+-----------------------------------------------------------+
    | *"editable"* (optional)             | Enable or disable editing in the column. If               |
    |                                     | nothing is given, it will follow the                      |
    |                                     | editability of the rest of the list.                      |
    +-------------------------------------+-----------------------------------------------------------+
    | *"width"* (optional)                | The width of the column.                                  |
    +-------------------------------------+-----------------------------------------------------------+
    | *"minWidth"* (optional)             | The minimum width of the column. The fallback is `width`. |
    +-------------------------------------+-----------------------------------------------------------+
    | *"maxWidth"* (optional)             | The maximum width of the column. The fallback is `width`. |
    +-------------------------------------+-----------------------------------------------------------+
    | *"sortable"* (optional)             | A boolean representing that this column allows the user   |
    |                                     | to sort the table by clicking the column's header.        |
    |                                     | The fallback is `True`. If a List is set to disallow      |
    |                                     | sorting the column level settings will be ignored.        |
    +-------------------------------------+-----------------------------------------------------------+
    | property (optional)                 | A property name for getting and setting the item value.   |
    +-------------------------------------+-----------------------------------------------------------+
    | getMethod (optional)                | A method name for getting the item value.                 |
    +-------------------------------------+-----------------------------------------------------------+
    | setMethod (optional)                | A method name for setting the item value.                 |
    +-------------------------------------+-----------------------------------------------------------+
    | getFunction (optional)              | A function for getting the item value.                    |
    +-------------------------------------+-----------------------------------------------------------+
    | setFunction (optional)              | A function for getting the item value.                    |
    +-------------------------------------+-----------------------------------------------------------+

    **showColumnTitles** Boolean representing if the column titles should be shown or not.
    Column titles will not be shown in single column lists.

    **selectionCallback** Callback to be called when the selection in the list changes.

    **doubleClickCallback** Callback to be called when an item is double clicked.

    **editCallback** Callback to be called after an item has been edited.

    # **menuCallback** Callback to be called when a contextual menu is requested.

    # **enableDelete** A boolean representing if items in the list can be deleted via the interface.

    **enableTypingSensitivity** A boolean representing if typing in the list will jump to the
    closest match as the entered keystrokes.

    **allowsSelection** A boolean representing if items in the list can be selected.

    **allowsMultipleSelection** A boolean representing if the list allows more than one item to be selected.

    **allowsEmptySelection** A boolean representing if the list allows zero items to be selected.

    **allowsSorting** A boolean indicating if the list allows user sorting by clicking column headers.

    **allowsColumnReordering** A boolean indicating if the list allows the user to reorder columns.

    **drawVerticalLines** Boolean representing if vertical lines should be drawn in the list.

    **drawHorizontalLines** Boolean representing if horizontal lines should be drawn in the list.

    **drawFocusRing** Boolean representing if the standard focus ring should be drawn when the list is selected.

    **alternatingRowColors** Boolean representing if alternating row colors should be used.

    **autohidesScrollers** Boolean representing if scrollbars should automatically be hidden if possible.

    Group Rows:

    It is possible to have rows that act as headers for a group of rows. To do this,
    add an instance of `List2GroupRow` to your items.::

        class Demo:

            def __init__(self):
                self.w = vanilla.Window((300, 150))
                items = [
                    vanilla.List2GroupRow("Group 1"),
                    "A",
                    "B",
                    "C",
                    vanilla.List2GroupRow("Group 2"),
                    "D",
                    "E",
                    "F"
                ]
                self.w.list = vanilla.List2(
                    "auto",
                    items=items,
                    allowsGroupRows=True,
                    floatsGroupRows=True,
                    allowsSorting=False
                )
                rules = [
                    "H:|[list]|",
                    "V:|[list]|"
                ]
                self.w.addAutoPosSizeRules(rules)
                self.w.open()

    **allowsGroupRows** Boolean representing if the list allows allows group rows.

    **floatsGroupRows** Boolean representing if the list floats the group rows.

    **groupRowCellClass**  A cell class to be displayed in the column.
    If nothing is given, a text cell is used.

    **groupRowCellClassArguments** A dictionary of keyword arguments to be used
    when *groupRowCellClass* is instantiated.

    **autosaveName** A string representing a unique name for the list. If given,
    this name will be used to store the column states in the application preferences.

    **dropSettings** A drop settings dictionary.

    Differences from the standard vanilla drag and drop API:

    Only these callbacks will be used:

    - `dropCandidateCallback`
    - `performDropCallback`
    - `dropCandidateEnteredCallback`
    - `dropCandidateEndedCallback`
    - `dropCandidateExitedCallback`

    `dropCandidateCallback` should return a boolean indicating if the
    drop is acceptable instead of a drop operation.

    The dragging info dictionary will contain an `index` key that specifies
    where in the list the is proposed for insertion. If the list doesn't
    allow dropping on or between rows, index will be `None`.
    """

    nsTableViewClass = VanillaList2TableViewSubclass
    dataSourceAndDelegateClass = VanillaList2DataSourceAndDelegate

    def __init__(self,
            posSize,
            items=[],
            columnDescriptions=[],
            allowsSelection=True,
            allowsMultipleSelection=True,
            allowsEmptySelection=True,
            allowsSorting=True,
            allowColumnReordering=True,
            enableTypingSensitivity=False,
            showColumnTitles=True,
            drawFocusRing=True,
            drawVerticalLines=False,
            drawHorizontalLines=False,
            alternatingRowColors=True,
            autohidesScrollers=False,
            selectionCallback=None,
            doubleClickCallback=None,
            editCallback=None,
            allowsGroupRows=False,
            floatsGroupRows=False,
            groupRowCellClass=None,
            groupRowCellClassArguments={},
            autosaveName=None,
            dragSettings=None,
            dropSettings=None
        ):
        if not columnDescriptions:
            showColumnTitles = False
            columnDescriptions = [
                dict(
                    identifier="value",
                    cellClass=EditTextList2Cell
                )
            ]
        self._tableView = getNSSubclass(self.nsTableViewClass)(self)
        self._dataSourceAndDelegate = self.dataSourceAndDelegateClass.alloc().initWithTableView_(self._tableView)
        self._tableView.setDataSource_(self._dataSourceAndDelegate)
        self._tableView.setDelegate_(self._dataSourceAndDelegate)
        # group rows
        if allowsGroupRows:
            assert not allowsSorting, "Group rows are not allowed in sortable lists."
            if groupRowCellClass is None:
                groupRowCellClass = GroupTitleList2Cell
            self._dataSourceAndDelegate.setGroupCellClassWithKwargs(groupRowCellClass, groupRowCellClassArguments)
            self._tableView.setFloatsGroupRows_(floatsGroupRows)
        # callbacks
        self._selectionCallback = selectionCallback
        self._editCallback = editCallback
        if doubleClickCallback is not None:
            self._doubleClickTarget = VanillaCallbackWrapper(doubleClickCallback)
            self._tableView.setTarget_(self._doubleClickTarget)
            self._tableView.setDoubleAction_("action:")
        # behavior attributes
        self._allowsSelection = allowsSelection
        self._tableView.setAllowsEmptySelection_(allowsEmptySelection)
        self._tableView.setAllowsMultipleSelection_(allowsMultipleSelection)
        self._tableView.setAllowsColumnReordering_(allowColumnReordering)
        self._allowsSorting = allowsSorting
        self._tableView.setAllowsTypeSelect_(enableTypingSensitivity)
        # visual attributes
        if not showColumnTitles:
            self._tableView.setHeaderView_(None)
            self._tableView.setCornerView_(None)
        self._tableView.setUsesAlternatingRowBackgroundColors_(alternatingRowColors)
        if not drawFocusRing:
            self._tableView.setFocusRingType_(AppKit.NSFocusRingTypeNone)
        if drawVerticalLines or drawHorizontalLines:
            if drawVerticalLines and drawHorizontalLines:
                lineType = AppKit.NSTableViewSolidVerticalGridLineMask | AppKit.NSTableViewSolidHorizontalGridLineMask
            elif drawVerticalLines:
                lineType = AppKit.NSTableViewSolidVerticalGridLineMask
            else:
                lineType = AppKit.NSTableViewSolidHorizontalGridLineMask
            self._tableView.setGridStyleMask_(lineType)
        if osVersionCurrent >= osVersion10_16:
            self._tableView.setStyle_(AppKit.NSTableViewStyleInset)
        # auto save
        if autosaveName is not None:
            self._tableView.setAutosaveName_(autosaveName)
            self._tableView.setAutosaveTableColumns_(True)
        # columns
        self._buildColumns(columnDescriptions)
        # self._tableView.setRowSizeStyle_(AppKit.NSTableViewRowSizeStyleDefault)
        super().__init__(
            posSize,
            nsView=self._tableView,
            autohidesScrollers=autohidesScrollers
        )
        # drag and drop
        if dragSettings is not None:
            self._dragCandidateCallback = dragSettings.get("dragCandidateCallback")
            self._dragStartedCallback = dragSettings.get("dragStartedCallback")
            self._makeDragDataCallback = dragSettings.get("makeDragDataCallback")
            self._dragEndedCallback = dragSettings.get("dragEndedCallback")
        if dropSettings is not None:
            self.setDropSettings(dropSettings)
        # populate
        self._itemsWereDict = True
        self.set(items)

    def _breakCycles(self):
        super()._breakCycles()
        self._selectionCallback = None
        self._editCallback = None

    def _getDropView(self):
        return self._tableView

    def _buildColumns(self, columnDescriptions):
        getters = {}
        setters = {}
        cellToValueConverters = {}
        valueToCellConverters = {}
        rowHeights = []
        if osVersionCurrent >= osVersion10_16:
            rowHeights.append(24)
        else:
            rowHeights.append(17)
        for columnDescription in columnDescriptions:
            identifier = columnDescription["identifier"]
            title = columnDescription.get("title", "")
            width = columnDescription.get("width")
            minWidth = columnDescription.get("minWidth", width)
            maxWidth = columnDescription.get("maxWidth", width)
            sortable = columnDescription.get("sortable")
            cellClass = columnDescription.get("cellClass", EditTextList2Cell)
            cellKwargs = columnDescription.get("cellClassArguments", {})
            editable = columnDescription.get("editable", False)
            property = columnDescription.get("property")
            getMethod = columnDescription.get("getMethod")
            setMethod = columnDescription.get("setMethod")
            getFunction = columnDescription.get("getFunction")
            setFunction = columnDescription.get("setFunction")
            cellToValueConverter = columnDescription.get("cellToValueConverter")
            valueToCellConverter = columnDescription.get("valueToCellConverter")
            getters[identifier] = dict(
                property=property,
                method=getMethod,
                function=getFunction
            )
            setters[identifier] = dict(
                property=property,
                method=setMethod,
                function=setFunction
            )
            if cellToValueConverter is not None:
                cellToValueConverters[identifier] = cellToValueConverter
            if valueToCellConverter is not None:
                valueToCellConverters[identifier] = valueToCellConverter
            cellKwargs["editable"] = editable
            if editable:
                cellKwargs["callback"] = True
            self._dataSourceAndDelegate.setCellClassWithKwargsForColumn(
                cellClass, cellKwargs, identifier
            )
            if width is not None:
                if width == minWidth and width == maxWidth:
                    resizingMask = AppKit.NSTableColumnNoResizing
                else:
                    resizingMask = AppKit.NSTableColumnUserResizingMask | AppKit.NSTableColumnAutoresizingMask
            else:
                resizingMask = AppKit.NSTableColumnUserResizingMask | AppKit.NSTableColumnAutoresizingMask
            column = AppKit.NSTableColumn.alloc().initWithIdentifier_(identifier)
            column.setTitle_(title)
            column.setResizingMask_(resizingMask)
            if width is not None:
                column.setWidth_(width)
                column.setMinWidth_(minWidth)
                column.setMaxWidth_(maxWidth)
            if self._allowsSorting and sortable:
                sortDescriptor = AppKit.NSSortDescriptor.sortDescriptorWithKey_ascending_selector_(
                    identifier,
                    True,
                    "compare:"
                )
                column.setSortDescriptorPrototype_(sortDescriptor)
            self._tableView.addTableColumn_(column)
            # measure the cell to get the row height
            cell = cellClass(**cellKwargs)
            height = cell._nsObject.fittingSize().height
            del cell
            rowHeights.append(height)
        self._dataSourceAndDelegate.setGetters(getters)
        self._dataSourceAndDelegate.setSetters(setters)
        self._dataSourceAndDelegate.setCellToValueConverters(cellToValueConverters)
        self._dataSourceAndDelegate.setValueToCellConverters(valueToCellConverters)
        self._tableView.setRowHeight_(max(rowHeights))

    def _wrapItem(self, item):
        if isinstance(item, simpleDataTypes):
            self._itemsWereDict = False
            item = dict(value=item)
        return item

    def getNSTableView(self):
        """
        Return the `NSTableView`_ that this object wraps.

        .. _NSTableView: https://developer.apple.com/documentation/appkit/nstableview?language=objc
        """
        return self._tableView

    def enable(self, onOff):
        """
        Enable or disable the object. **onOff** should be a boolean.
        """
        self._tableView.setEnabled_(onOff)

    # Data

    def set(self, items):
        """
        Set the items in the list.

        **items** should follow the same format as described in the constructor.
        """
        items = [self._wrapItem(item) for item in items]
        self._dataSourceAndDelegate.setItems(items)

    def get(self):
        """
        Get the list of items in the list.
        """
        items = list(self._dataSourceAndDelegate.items())
        if not self._itemsWereDict:
            items = [item["value"] for item in items]
        return items

    def getArrangedIndexes(self):
        """
        Get the indexes of the items as they appear
        to the user in the list.
        """
        return self._dataSourceAndDelegate.arrangedIndexes()

    def getArrangedItems(self):
        """
        Get the items as they appear to the user in the list.
        """
        return self._dataSourceAndDelegate.arrangedItems()

    def reloadData(self, indexes=None):
        """
        Reload the data on display. This is needed when
        the items have changed values and the presentation
        of the items in the list needs to be updated.

        **indexes** may be provided to indicate that only
        specific items need to be reloaded.
        """
        tableView = self._tableView
        if indexes is None:
            tableView.reloadData()
        else:
            rowIndexes = makeIndexSet(indexes)
            columnIndexes = range(len(tableView.tableColumns()))
            columnIndexes = makeIndexSet(columnIndexes)
            tableView.reloadDataForRowIndexes_columnIndexes_(
                rowIndexes,
                columnIndexes
            )

    # Selection

    def getSelectedItems(self):
        """
        Get a list of selected items in the list.
        """
        selectedIndexes = self._tableView.selectedRowIndexes()
        items = self.get()
        indexes = self.getArrangedIndexes()
        items = [
            items[i]
            for i in indexes
            if i in selectedIndexes
        ]
        return items

    def setSelectedItems(self, items):
        """
        Set the selected items in the list.

        .. note::
           `setSelectedIndexes` is the recommended method
           for setting selection. `setSelectedItems` is
           a convenience method that relies on iteration
           and comparing object ids to find the item indexes,
           which are then sent to `setSelectedIndexes`. If
           the list contains a large number of items this
           iteration can create a performance problem.
        """
        allItems = self.get()
        indexMapping = {
            id(item) : index
            for index, item in enumerate(allItems)
        }
        selectionIndexes = [
            indexMapping[id(item)]
            for item in items
        ]
        self.setSelectedIndexes(selectionIndexes)

    def getSelectedIndexes(self):
        """
        Get a list of indexes of selected items in the list.
        """
        rowIndexes = self._tableView.selectedRowIndexes()
        arrangedIndexes = self.getArrangedIndexes()
        itemIndexes = [
            arrangedIndexes[i]
            for i in rowIndexes
        ]
        return itemIndexes

    def setSelectedIndexes(self, indexes):
        """
        Set the selected indexes in the list.

        **indexes** should be a list of indexes.
        """
        itemIndexToRowIndexes = {
            itemIndex : rowIndex
            for rowIndex, itemIndex
            in enumerate(self.getArrangedIndexes())
        }
        rowIndexes = [
            itemIndexToRowIndexes[itemIndex]
            for itemIndex in indexes
        ]
        rowIndexes = makeIndexSet(rowIndexes)
        self._tableView.selectRowIndexes_byExtendingSelection_(rowIndexes, False)

    def scrollToSelection(self):
        """
        Scroll the selected rows to visible.
        """
        selection = self.getSelectedIndexes()
        if not selection:
            return
        index = min(selection)
        self.scrollToIndex(index)

    def scrollToIndex(self, row):
        """
        Scroll the row to visible.
        """
        self._tableView.scrollRowToVisible_(row)

    # Drag

    _dragCandidateCallback = None
    _dragStartedCallback = None
    _makeDragDataCallback = None
    _dragEndedCallback = None

    def _validateRowsForDrag(self, indexes):
        if self._dragCandidateCallback is None:
            if self._makeDragDataCallback is None:
                return False
            return True
        return self._dragCandidateCallback(indexes)

    def _getPasteboardDataForIndex(self, index):
        if self._makeDragDataCallback is None:
            return None
        typesAndValues = self._makeDragDataCallback(index)
        if not typesAndValues:
            return None
        return makePasteboardItem(typesAndValues)

    # Drop

    _allowDropOnRow = None
    _allowDropBetweenRows = None

    def setDropSettings(self, settings):
        self._allowDropOnRow = settings.get("allowDropOnRow", False)
        self._allowDropBetweenRows = settings.get("allowDropBetweenRows", True)
        super().setDropSettings(settings)

    def _dropCandidateUpdated(self, draggingInfo, index, operation):
        if self._dropCandidateCallback is None:
            return AppKit.NSDragOperationNone
        info = self._unpackDropCandidateInfo(draggingInfo)
        info["index"] = index
        if not self._allowDropOnRow and not self._allowDropBetweenRows:
            index = None
        elif not self._allowDropOnRow and operation == AppKit.NSTableViewDropOn:
            return AppKit.NSDragOperationNone
        elif not self._allowDropBetweenRows and operation == AppKit.NSTableViewDropAbove:
            return AppKit.NSDragOperationNone
        if not len(self.get()):
            index = None
        operation = self._dropCandidateCallback(info)
        operation = dropOperationMap.get(operation, operation)
        # highlight the whole table instead of a single spot
        if index is None:
            self._tableView.setDropRow_dropOperation_(-1, operation)
        return operation

    def _performDrop(self, draggingInfo, index, operation):
        if not self._allowDropOnRow and not self._allowDropBetweenRows:
            index = None
        info = self._unpackDropCandidateInfo(draggingInfo)
        info["index"] = index
        return self._performDropCallback(info)

# -----
# Tools
# -----

class List2GroupRow:

    def __init__(self, value=None):
        self.value = value


def makeIndexSet(indexes):
    indexSet = AppKit.NSMutableIndexSet.indexSet()
    for i in indexes:
        indexSet.addIndex_(i)
    return indexSet

# -----
# Cells
# -----

from vanilla.vanillaGroup import Group
from vanilla.vanillaEditText import EditText

class EditTextList2Cell(Group):

    """
    An object that displays text in a List2 column.

    **verticalAlignment** The vertical alignment of the text
    within the row. Options:

    - `"top"`
    - `"center"`
    - `"bottom"`

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    # Implementation Note:
    # Using NSTextField directly as the row view doesn't
    # allow for vertically positioning the text anywhere
    # but to the top. In macOS 11 this leads to text
    # being out of alignment with other cell types like
    # checkboxes and sliders. I found numerous complaints
    # about this on Stack Overflow so I'm not alone in
    # noticing this. This is apparently solved automatically
    # when using IB, but a pain when building things
    # programatically. Additionally, I chose to use NSView
    # as the base of this instead of NSTableCellView because
    # NSTableCellView doesn't have a textField by default
    # so it has to be built and we may as well skip the
    # expense of NSTableCellView and go straight to NSView.
    # Sigh.

    def __init__(self,
            verticalAlignment="center",
            editable=False,
            callback=None
        ):
        self._externalCallback = callback
        super().__init__("auto")
        self.editText = EditText(
            "auto",
            readOnly=not editable,
            callback=self._internalCallback
        )
        container = self._nsObject
        textField = self.editText.getNSTextField()
        if verticalAlignment == "top":
            yRule = "V:|[editText]"
        elif verticalAlignment == "bottom":
            yRule = "V:[editText]|"
        else:
            yRule = dict(
                view1=self.editText,
                attribute1="centerY",
                view2=self,
                attribute2="centerY"
            )
        rules = [
            "H:|[editText]|",
            yRule
        ]
        if verticalAlignment != "center":
            heightRule = dict(
                view1=self.editText,
                attribute1="height",
                view2=self,
                attribute2="height"
            )
            rules.append(heightRule)
        self.addAutoPosSizeRules(rules)
        textField.setDrawsBackground_(False)
        textField.setBezeled_(False)

    def getNSTextField(self):
        return self.editText.getNSTextField()

    def _internalCallback(self, sender):
        if self._externalCallback is not None:
            self._externalCallback(self)

    def set(self, value):
        self.editText.set(value)

    def get(self):
        return self.editText.get()


class GroupTitleList2Cell(EditTextList2Cell):

    """
    An object that displays highlighted text for group rows
    in a List2 column.

    **font** The font.

    **textColor** The text color.

    **backgroundColor** The background color.

    **verticalAlignment** The vertical alignment of the text
    within the row. Options:

    - `"top"`
    - `"center"`
    - `"bottom"`

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            font=None,
            textColor=None,
            backgroundColor=None,
            verticalAlignment="center",
            editable=False,
            callback=None
        ):
        super().__init__(
            editable=False,
            callback=callback
        )
        textField = self.getNSTextField()
        if backgroundColor is not None:
            textField.setDrawsBackground_(True)
            textField.setBackgroundColor_(backgroundColor)
        if textColor is not None:
            textField.setTextColor_(textColor)
        if font is not None:
            textField.setFont_(font)



from vanilla.vanillaSlider import Slider

class SliderList2Cell(Slider):

    """
    An object that displays a slider in a List2 column.

    **minValue** The minimum value for the slider.

    **maxValue** The maximum value for the slider.

    **tickMarkCount** The number of tick marks to be displayed on the slider.
    If *None* is given, no tick marks will be displayed.

    **stopOnTickMarks** Boolean representing if the slider knob should only
    stop on the tick marks.

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            minValue=0,
            maxValue=100,
            value=50,
            tickMarkCount=0,
            stopOnTickMarks=False,
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            minValue=minValue,
            maxValue=maxValue,
            value=value,
            tickMarkCount=tickMarkCount,
            stopOnTickMarks=stopOnTickMarks,
            sizeStyle="small",
            callback=callback
        )
        self.enable(editable)


from vanilla.vanillaCheckBox import CheckBox

class CheckBoxList2Cell(CheckBox):

    """
    An object that displays a check box in a List2 column.

    **title** The title to be set in *all* items in the List column.

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            title=None,
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            title=title,
            sizeStyle="small",
            callback=callback
        )
        self.enable(editable)


from vanilla.vanillaPopUpButton import PopUpButton

class PopUpButtonList2Cell(PopUpButton):

    """
    An object that displays a pop up list in a List2 column.

    **items** The items that should appear in the pop up list.

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            items=[],
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            items=items,
            sizeStyle="small",
            callback=callback
        )
        self.enable(editable)


from vanilla.vanillaImageView import ImageView

class ImageList2Cell(ImageView):

    """
    An object that displays an image in a List2 column.

    **horizontalAlignment** A string representing the desired horizontal
    alignment of the image in the view. The options are:

    +-------------+-------------------------+
    | "left"      | Image is aligned left.  |
    +-------------+-------------------------+
    | "right"     | Image is aligned right. |
    +-------------+-------------------------+
    | "center"    | Image is centered.      |
    +-------------+-------------------------+

    **verticalAlignment** A string representing the desired vertical alignment
    of the image in the view. The options are:

    +-------------+--------------------------+
    | "top"       | Image is aligned top.    |
    +-------------+--------------------------+
    | "bottom"    | Image is aligned bottom. |
    +-------------+--------------------------+
    | "center"    | Image is centered.       |
    +-------------+--------------------------+

    **scale** A string representing the desired scale style of the image in the
    view. The options are:

    +----------------+----------------------------------------------+
    | "proportional" | Proportionally scale the image to fit in the |
    |                | view if it is larger than the view.          |
    +----------------+----------------------------------------------+
    | "fit"          | Distort the proportions of the image until   |
    |                | it fits exactly in the view.                 |
    +----------------+----------------------------------------------+
    | "none"         | Do not scale the image.                      |
    +----------------+----------------------------------------------+

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            horizontalAlignment="center",
            verticalAlignment="center",
            scale="proportional",
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            horizontalAlignment=horizontalAlignment,
            verticalAlignment=verticalAlignment,
            scale=scale
        )

    def set(self, image):
        self.setImage(imageObject=image)

    def get(self):
        return self.getNSImageView().image()


from vanilla.vanillaSegmentedButton import SegmentedButton

class SegmentedButtonList2Cell(SegmentedButton):

    """
    An object that displays a segmented button in a List2 column.

    **segmentDescriptions** An ordered list of dictionaries describing the segments.

    +------------------------+-----------------------------------------------------+
    | title (optional)       | The title of the segment.                           |
    +------------------------+-----------------------------------------------------+
    | imagePath (optional)   | A file path to an image to display in the segment.  |
    +------------------------+-----------------------------------------------------+
    | imageNamed (optional)  | The name of an image already loaded as a `NSImage`_ |
    |                        | by the application to display in the segment.       |
    +------------------------+-----------------------------------------------------+
    | imageObject (optional) | A `NSImage`_ object to display in the segment.      |
    +------------------------+-----------------------------------------------------+

    .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            segmentDescriptions=[],
            selectionStyle="one",
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            segmentDescriptions=segmentDescriptions,
            selectionStyle=selectionStyle,
            sizeStyle="small",
            callback=callback
        )
        self.enable(editable)


from vanilla.vanillaColorWell import ColorWell

class ColorWellList2Cell(ColorWell):

    """
    An object that displays a color well in a List2 column.

    **title** The title to be set in *all* items in the List column.

    .. note::
       This class should only be used in the *columnDescriptions*
       *cellClass* argument during the construction of a List.
       This is never constructed directly.
    """

    def __init__(self,
            editable=False,
            callback=None
        ):
        super().__init__(
            "auto",
            callback=callback
        )
        colorWell = self.getNSColorWell()
        self.enable(editable)


from vanilla import *

class TinyTextEditorDocumentWindow(object):
    
    def __init__(self):
        self.w = Window((400, 400), minSize=(100, 100))
        self.w.textEditor = TextEditor((0, 0, -0, -0), checksSpelling=True)
        self.w.open()
    
    def getText(self):
        return self.w.textEditor.get()
    
    def setText(self, text):
        self.w.textEditor.set(text)
    
    def assignToDocument(self, nsDocument):
        self.w.assignToDocument(nsDocument)



# from Foundation import *
from AppKit import NSDate, NSDatePicker, NSClockAndCalendarDatePickerStyle, NSTextFieldAndStepperDatePickerStyle, NSHourMinuteDatePickerElementFlag, NSHourMinuteSecondDatePickerElementFlag, NSYearMonthDatePickerElementFlag, NSYearMonthDayDatePickerElementFlag
from vanilla.vanillaBase import VanillaBaseControl


_modeMap = {
    "graphical": NSClockAndCalendarDatePickerStyle,
    "text": 2, #NSTextFieldDatePickerStyle raises an error
    "textStepper": NSTextFieldAndStepperDatePickerStyle,
}

_timeDisplayFlagMap = {
    None : 0,
    "hourMinute" : NSHourMinuteDatePickerElementFlag,
    "hourMinuteSecond" : NSHourMinuteSecondDatePickerElementFlag,
}

_dateDisplayFlagMap = {
    None : 0,
    "yearMonth" : NSYearMonthDatePickerElementFlag,
    "yearMonthDay" : NSYearMonthDayDatePickerElementFlag,
}


class DatePicker(VanillaBaseControl):

    """
    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the date picker control.

    +-------------------------------------+
    | **Standard Dimensions - Text Mode** |
    +---------+---+-----------------------+
    | Regular | H | 22                    |
    +---------+---+-----------------------+
    | Small   | H | 19                    |
    +---------+---+-----------------------+
    | Mini    | H | 16                    |
    +---------+---+-----------------------+

    +------------------------------------------+
    | **Standard Dimensions - Graphical Mode** |
    +--------------------+---------------------+
    | Calendar and Clock | 227w 148h           |
    +--------------------+---------------------+
    | Calendar           | 139w 148h           |
    +--------------------+---------------------+
    | Clock              | 122w 123h           |
    +--------------------+---------------------+

    **date** A `NSDate`_ object representing the date and time that should be
    set in the control.

    **minDate** A `NSDate`_ object representing the lowest date and time that
    can be set in the control.

    **maxDate** A `NSDate`_ object representing the highest date and time that
    can be set in the control.

    **showStepper** A boolean indicating if the thumb stepper should be shown
    in text mode.

    **mode** A string representing the desired mode for the date picker control.
    The options are:

    +-------------+
    | "text"      |
    +-------------+
    | "graphical" |
    +-------------+

    **timeDisplay** A string representing the desired time units that should be
    displayed in the date picker control. The options are:

    +--------------------+-------------------------------+
    | None               | Do not display time.          |
    +--------------------+-------------------------------+
    | "hourMinute"       | Display hour and minute.      |
    +--------------------+-------------------------------+
    | "hourMinuteSecond" | Display hour, minute, second. |
    +--------------------+-------------------------------+

    **dateDisplay** A string representing the desired date units that should be
    displayed in the date picker control. The options are:

    +----------------+------------------------------+
    | None           | Do not display date.         |
    +----------------+------------------------------+
    | "yearMonth"    | Display year and month.      |
    +----------------+------------------------------+
    | "yearMonthDay" | Display year, month and day. |
    +----------------+------------------------------+

    **sizeStyle** A string representing the desired size style of the
    date picker control. This only applies in text mode. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSDate: https://developer.apple.com/documentation/foundation/nsdate?language=objc
    """

    nsDatePickerClass = NSDatePicker

    def __init__(self, posSize, date=None, minDate=None, maxDate=None, showStepper=True, mode="text",
        timeDisplay="hourMinuteSecond", dateDisplay="yearMonthDay", callback=None, sizeStyle="regular"):
        self._setupView(self.nsDatePickerClass, posSize, callback=callback)
        self._setSizeStyle(sizeStyle)
        self._nsObject.setDrawsBackground_(True)
        self._nsObject.setBezeled_(True)
        if mode == "text" and showStepper:
            mode += "Stepper"
        style = _modeMap[mode]
        self._nsObject.setDatePickerStyle_(style)
        flag = _timeDisplayFlagMap[timeDisplay]
        flag = flag | _dateDisplayFlagMap[dateDisplay]
        self._nsObject.setDatePickerElements_(flag)
        if date is None:
            date = NSDate.date()
        self._nsObject.setDateValue_(date)
        if minDate is not None:
            self._nsObject.setMinDate_(minDate)
        if maxDate is not None:
            self._nsObject.setMaxDate_(maxDate)

    def getNSDatePicker(self):
        """
        Return the `NSDatePicker`_ that this object wraps.

        .. _NSDatePicker: https://developer.apple.com/documentation/appkit/nsdatepicker?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the contents of the date picker control.
        """
        return self._nsObject.dateValue()

    def set(self, value):
        """
        Set the contents of the date picker control.

        **value** A `NSDate`_ object.
        """
        self._nsObject.setDateValue_(value)



from __future__ import print_function
import sys
from AppKit import NSObject
from PyObjCTools import AppHelper
import vanilla


class SimpleAppAppDelegate(NSObject):

    def applicationDidFinishLaunching_(self, notification):
        SimpleAppWindow()


class SimpleAppWindow(object):

    def __init__(self):
        self.w = vanilla.Window((250, 120), "Simple App Window", closable=False)
        self.w.text = vanilla.TextBox((10, 10, -10, 70), "This is a simple window. It doesn't do much. You see that button? Press it and some text will be printed in Console.app.")
        self.w.button = vanilla.Button((10, 90, -10, 20), "Press me", callback=self.buttonCallback)
        self.w.open()

    def buttonCallback(self, sender):
        print("You pressed the button!")
        sys.stdout.flush()


if __name__ == "__main__":
    AppHelper.runEventLoop()




from AppKit import NSColorWell, NSColorPanel
from vanilla.vanillaBase import VanillaBaseObject


class ColorWell(VanillaBaseObject):

    """
    A control that allows for showing and choosing a color value.

    .. image:: /_images/ColorWell.png

    ColorWell objects handle `NSColor`_ objects.

    ::

        from AppKit import NSColor
        from vanilla import Window, ColorWell

        class ColorWellDemo:

            def __init__(self):
                self.w = Window((100, 50))
                self.w.colorWell = ColorWell((10, 10, -10, -10),
                                    callback=self.colorWellEdit,
                                    color=NSColor.redColor())
                self.w.open()

            def colorWellEdit(self, sender):
                print("color well edit!", sender.get())

        ColorWellDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the color well.

    **callback** The method to be caled when the user selects a new color.

    **color** A `NSColor`_ object. If *None* is given, the color shown will
    be white.

    .. _NSColor: https://developer.apple.com/documentation/appkit/nscolor?language=objc
    """

    nsColorWellClass = NSColorWell

    def __init__(self, posSize, callback=None, color=None):
        self._setupView(self.nsColorWellClass, posSize, callback=callback)
        if color is not None:
            self._nsObject.setColor_(color)
        colorPanel = NSColorPanel.sharedColorPanel()
        colorPanel.setShowsAlpha_(True)

    def getNSColorWell(self):
        """
        Return the `NSColorWell`_ that this object wraps.

        .. _NSColorWell: https://developer.apple.com/documentation/appkit/nscolorwell?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the `NSColor`_ object representing the current color in
        the color well.

        .. _NSColor: https://developer.apple.com/documentation/appkit/nscolor?language=objc
        """
        return self._nsObject.color()

    def set(self, color):
        """
        Set the color in the color well.

        **color** A `NSColor`_ object representing the color to be displayed
        in the color well.
        """
        return self._nsObject.setColor_(color)



from vanilla.vanillaBase import VanillaBaseObject, VanillaBaseControl, VanillaError
from vanilla.vanillaBox import Box, HorizontalLine, VerticalLine
from vanilla.vanillaBrowser import ObjectBrowser
from vanilla.vanillaButton import Button, SquareButton, ImageButton, HelpButton
from vanilla.vanillaCheckBox import CheckBox
from vanilla.vanillaColorWell import ColorWell
from vanilla.vanillaComboBox import ComboBox
from vanilla.vanillaDatePicker import DatePicker
from vanilla.vanillaDrawer import Drawer
from vanilla.vanillaEditText import EditText, SecureEditText
from vanilla.vanillaGradientButton import GradientButton
from vanilla.vanillaGroup import Group
from vanilla.vanillaImageView import ImageView
from vanilla.vanillaLevelIndicator import LevelIndicator, LevelIndicatorListCell
from vanilla.vanillaList import List, CheckBoxListCell, SliderListCell, PopUpButtonListCell, ImageListCell, SegmentedButtonListCell
from vanilla.vanillaList2 import List2, List2GroupRow, EditTextList2Cell, GroupTitleList2Cell, SliderList2Cell, CheckBoxList2Cell, PopUpButtonList2Cell, ImageList2Cell, SegmentedButtonList2Cell, ColorWellList2Cell
from vanilla.vanillaPathControl import PathControl
from vanilla.vanillaPopUpButton import PopUpButton, ActionButton
from vanilla.vanillaPopover import Popover
from vanilla.vanillaProgressBar import ProgressBar
from vanilla.vanillaProgressSpinner import ProgressSpinner
from vanilla.vanillaRadioGroup import RadioGroup, VerticalRadioGroup, HorizontalRadioGroup, RadioButton
from vanilla.vanillaScrollView import ScrollView
from vanilla.vanillaSearchBox import SearchBox
from vanilla.vanillaSegmentedButton import SegmentedButton
from vanilla.vanillaSlider import Slider
from vanilla.vanillaSplitView import SplitView, SplitView2
from vanilla.vanillaStackGroup import HorizontalStackGroup, VerticalStackGroup
from vanilla.vanillaStackView import HorizontalStackView, VerticalStackView
from vanilla.vanillaTabs import Tabs
from vanilla.vanillaTextBox import TextBox
from vanilla.vanillaTextEditor import TextEditor
from vanilla.vanillaWindows import Window, FloatingWindow, HUDFloatingWindow, Sheet
from vanilla.dragAndDrop import startDraggingSession, DropTargetProtocolMixIn

__all__ = [
    "VanillaBaseObject", "VanillaBaseControl", "VanillaError",
    "Box", "HorizontalLine", "VerticalLine",
    "Button", "SquareButton", "ImageButton", "HelpButton",
    "CheckBox",
    "ColorWell",
    "ComboBox",
    "DatePicker",
    "Drawer",
    "EditText",
    "GradientButton",
    "Group",
    "ImageView",
    "LevelIndicator", "LevelIndicatorListCell",
    "List", "CheckBoxListCell", "SliderListCell", "PopUpButtonListCell", "ImageListCell", "SegmentedButtonListCell",
    "List2", "List2GroupRow", "EditTextList2Cell", "GroupTitleList2Cell", "SliderList2Cell", "CheckBoxList2Cell", "PopUpButtonList2Cell", "ImageList2Cell", "SegmentedButtonList2Cell", "ColorWellList2Cell",
    "ObjectBrowser",
    "PathControl",
    "PopUpButton", "ActionButton",
    "Popover",
    "ProgressBar",
    "ProgressSpinner",
    "RadioGroup", "VerticalRadioGroup", "HorizontalRadioGroup", "RadioButton",
    "ScrollView",
    "SearchBox",
    "SecureEditText",
    "SegmentedButton",
    "Slider",
    "SplitView",
    "SplitView2",
    "HorizontalStackGroup", "VerticalStackGroup",
    "HorizontalStackView", "VerticalStackView",
    "Tabs",
    "TextBox",
    "TextEditor",
    "Window", "FloatingWindow", "HUDFloatingWindow", "Sheet",

    "startDraggingSession",
    "DropTargetProtocolMixIn"
    ]


try:
    from ._version import version as __version__
except ImportError:
    __version__ = "<unknown>"


# NSGridview is available from OS 10.12+
try:
    from AppKit import NSGridView
except ImportError:
    pass
else:
    from vanilla.vanillaGridView import GridView
    __all__.append("GridView")



import warnings

warnings.warn(
    "The py23 module has been deprecated and will be removed. "
    "Please update your code.",
    DeprecationWarning,
)

__all__ = ['basestring', 'unicode', 'long', 'python_method']

try:
    basestring = basestring
except NameError:
    basestring = str

try:
    unicode = unicode
except NameError:
    unicode = str

try:
    long = long
except NameError:
    long = int

try:
    range = xrange
except NameError:
    range = range

try:
    unichr = unichr
except NameError:
    unichr = chr

try:
    from objc import python_method
except ImportError:
    def python_method(arg):
        return arg



import vanilla
from vanilla import dialogs


class Test(object):

    def __init__(self):
        buttons = [
            "message",
            "messageSheet",
            "messageIcon",
            "messageAccessoryView",
            "messageShowHelp",
            "messageCritical",

            "askYesNo",
            "askYesNoCallback",
            "askYesNoSheet",
            "askYesNoCancel",
            "askYesNoCancelCallback",
            "askYesNoCancelSheet",
            "askVanillaStrawberry",
            "askCustomCallback",

            "getFile",
            "getFileSheet",
            "getFolder",
            "getFolderSheet",
            "getFileOrFolder",
            "getFileOrFolderSheet",
            "getFileAccessoryView"
        ]

        self.w = vanilla.Window((200, (len(buttons) + 1) * 20))

        y = 10
        for b in buttons:
            obj = vanilla.Button((10, y, -10, 16), b, callback=getattr(self, b), sizeStyle="mini")
            setattr(self.w, "button_%s" % b, obj)
            y += 20

        self.w.open()

    def dummyCallback(self, value):
        print(value)

    def message(self, sender):
        dialogs.message("Hello world", "foo bar")

    def messageSheet(self, sender):
        dialogs.message("Hello world", "foo bar", parentWindow=self.w)

    def messageIcon(self, sender):
        import AppKit
        dialogs.message("Hello world", "Show folder icon", icon=AppKit.NSImage.imageNamed_(AppKit.NSImageNameFolder))

    def messageAccessoryView(self, sender):
        import vanilla
        view = vanilla.Box((0, 0, 200, 75))
        view.checkBox = vanilla.CheckBox((10, 10, 100, 22), "checked")
        dialogs.message("Hello world", "foo bar", accessoryView=view)

    def messageShowHelp(self, sender):
        def help():
            print("help")
        dialogs.message("Hello world", "foo bar", showsHelpCallback=help)

    def messageCritical(self, sender):
        dialogs.message("Hello world", "foo bar", alertStyle="critical")

    def askYesNo(self, sender):
        print("result", dialogs.askYesNo("Hello world", "foo bar"))

    def askYesNoCallback(self, sender):
        dialogs.askYesNo("Hello world", "foo bar", resultCallback=self.dummyCallback)

    def askYesNoSheet(self, sender):
        dialogs.askYesNo("Hello world", "foo bar", parentWindow=self.w, resultCallback=self.dummyCallback)

    def askYesNoCancel(self, sender):
        print("result", dialogs.askYesNoCancel("Hello world", "foo bar"))

    def askYesNoCancelCallback(self, sender):
        dialogs.askYesNoCancel("Hello world", "foo bar", resultCallback=self.dummyCallback)

    def askYesNoCancelSheet(self, sender):
        dialogs.askYesNoCancel("Hello world", "foo bar", parentWindow=self.w, resultCallback=self.dummyCallback)

    def askVanillaStrawberry(self, sender):
        dialogs.ask("Hello world", "foo bar", buttonTitles=[("vanilla", 1), ("Strawberry", 0)], parentWindow=self.w, resultCallback=self.dummyCallback)

    def askCustomCallback(self, sender):

        def go():
            print("go")
        def stop():
            print("stop")

        dialogs.ask("Hello world", "foo bar", buttonTitles=[dict(title="go", callback=go), dict(title="stop", callback=stop)], parentWindow=self.w, resultCallback=self.dummyCallback)

    def getFile(self, sender):
        print("result", dialogs.getFile())

    def getFileSheet(self, sender):
        dialogs.getFile(parentWindow=self.w, resultCallback=self.dummyCallback)

    def getFolder(self, sender):
        print("result", dialogs.getFolder())

    def getFolderSheet(self, sender):
        dialogs.getFolder(parentWindow=self.w, resultCallback=self.dummyCallback)

    def getFileOrFolder(self, sender):
        print("result", dialogs.getFileOrFolder())

    def getFileOrFolderSheet(self, sender):
        dialogs.getFileOrFolder(parentWindow=self.w, resultCallback=self.dummyCallback)

    def getFileAccessoryView(self, sender):
        import vanilla
        view = vanilla.Box((0, 0, 200, 75))
        view.checkBox = vanilla.CheckBox((10, 10, 100, 22), "checked")
        print("result", dialogs.getFile(accessoryView=view))


if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)


from AppKit import NSScrollView, NSBezelBorder
from vanilla.vanillaBase import VanillaBaseObject


class ScrollView(VanillaBaseObject):

    """
    A view with scrollers for containing another view.

    .. image:: /_images/ScrollView.png

    ::

        from AppKit import NSView, NSColor, NSRectFill
        from vanilla import Window, ScrollView

        class DemoView(NSView):

            def drawRect_(self, rect):
                NSColor.redColor().set()
                NSRectFill(self.bounds())

        class ScrollViewDemo:

            def __init__(self):
                self.w = Window((200, 200))
                self.view = DemoView.alloc().init()
                self.view.setFrame_(((0, 0), (300, 300)))
                self.w.scrollView = ScrollView((10, 10, -10, -10),
                                        self.view)
                self.w.open()

        ScrollViewDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the scroll view.

    **nsView** A `NSView`_ object.

    **hasHorizontalScroller** Boolean representing if the scroll view has horizontal scrollers.

    **hasVerticalScroller** Boolean representing if the scroll view has vertical scrollers.

    **autohidesScrollers** Boolean representing if the scroll view auto-hides its scrollers.

    **backgroundColor** A `NSColor`_ object representing the background color of the scroll view.

    **drawsBackground** Boolean representing if the background should be drawn.

    .. _NSView: https://developer.apple.com/documentation/appkit/nsview?language=objc
    .. _NSColor: https://developer.apple.com/documentation/appkit/nscolor?language=objc
    """

    nsScrollViewClass = NSScrollView

    def __init__(self, posSize, nsView, hasHorizontalScroller=True, hasVerticalScroller=True,
                    autohidesScrollers=False, backgroundColor=None, clipView=None, drawsBackground=True):
        self._setupView(self.nsScrollViewClass, posSize)
        if clipView is not None:
            self._nsObject.setContentView_(clipView)
        self._nsObject.setDocumentView_(nsView)
        self._nsObject.setHasHorizontalScroller_(hasHorizontalScroller)
        self._nsObject.setHasVerticalScroller_(hasVerticalScroller)
        self._nsObject.setAutohidesScrollers_(autohidesScrollers)
        self._nsObject.setBorderType_(NSBezelBorder)
        if backgroundColor:
            self._nsObject.setBackgroundColor_(backgroundColor)
        self._nsObject.setDrawsBackground_(drawsBackground)

    def _testForDeprecatedAttributes(self):
        super(ScrollView, self)._testForDeprecatedAttributes()
        from warnings import warn
        if hasattr(self, "_scrollViewClass"):
            warn(DeprecationWarning("The _scrollViewClass attribute is deprecated. Use the nsScrollViewClass attribute."))
            self.nsScrollViewClass = self._scrollViewClass

    def getNSScrollView(self):
        """
        Return the `NSScrollView`_ that this object wraps.

        .. _NSScrollView: https://developer.apple.com/documentation/appkit/nsscrollview?language=objc
        """
        return self._nsObject

    def setBackgroundColor(self, color):
        """
        Set the background of the scrol view to *color*.
        """
        self._nsObject.setBackgroundColor_(color)

